try:
    import configparser as ConfigParser
except:
    import ConfigParser
import io
import os
import errno

from six.moves import input


class ConnectionConfig(dict):
    """Loads or creates .cfg file to/from object

    Config properties can be accessed as attributes or as dictionary keys.
    Treating object as dict only contains config items, not special properties.
    """

    properties = ("path", "server")

    def __init__(self, server, path=None, **kwargs):
        """
        Args:
            server (str): Name of server section in config file
            path (str, optional): Path to config file
        """
        self.server = server

        self.path = path
        if path:
            self.load_from_file(path=path)

    def load_from_file(self, path=None, stream=None):
        """Loads config file

        Args:
            path (str, optional): Path to file. specifiying this also updates the
                `path` property of the ConnectionConfig instance.
            stream (optional): Text stream of config.
        """
        config = ConfigParser.ConfigParser()
        if path and stream:
            raise IOError("Cannot pass both path and stream")

        if not os.path.exists(path):
            with open(path, "w") as fid:
                pass

        if path:
            config.read_file(open(path))
        elif stream:
            config.read_file(io.BytesIO(stream))

        self.path = path
        self.update({t[0]: t[1] for t in config.items(self.server)})

        return self

    def save(self, path=None):
        """Saves section to config file

        Args:
            path (str, optional): Path to save file instead of object's
                 `path` property. Does NOT change `path` property.
        """
        path = path if path else self.path
        config = ConfigParser.ConfigParser()

        if path:
            try:
                config.readfp(open(path))
            except IOError as e:
                if not e.errno == errno.ENOENT:
                    raise

        try:
            config.add_section(self.server)
        except ConfigParser.DuplicateSectionError as e:
            pass

        # Set server properties
        config.set(self.server, "url", self.url)

        if self.auth_url:
            config.set(self.server, "auth_url", self.auth_url)

        config.write(open(path, "w"))

        return self

    def save_config(self, path=None):
        """Prompts for user input, if needed, before calling save()

        Args:
            path (str, optional): Save location to use instead of `path` property.
        """
        path = path if path else self.path

        self.url = self.url or input("Enter server URL: ")

        if self.url[-1] != "/":
            self.url += "/"

        self.auth_url = self.url + "oauth/"

        config = self.save(path)

        print(
            "Saved configuration file to {}".format(
                os.path.join(os.path.abspath("."), path)
            )
        )

        return self

    def get_server_dict(self):
        return self

    def create_config_object(self, path=None, url=None, auth_url=None):
        """Creates new config section if needed"""
        path = path if path else self.path

        self.url = url
        self.auth_url = auth_url

        if not path:
            path = input("Enter name of config file (connect.cfg): ") or "connect.cfg"
        self.save_config()

        return self

    def __getattr__(self, key):
        try:
            return self.__getitem__(key)
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, key, value):
        super(ConnectionConfig, self).__setattr__(key, value)
        if key in self.properties:
            return
        try:
            super(ConnectionConfig, self).__setitem__(key, value)
        except KeyError as e:
            raise AttributeError(e)

    def __setitem__(self, key, value):
        if key in self.properties:
            raise KeyError("This key is reserved by the system.")
        super(ConnectionConfig, self).__setitem__(key, value)
        try:
            super(ConnectionConfig, self).__setattr__(key, value)
        except AttributeError as e:
            raise KeyError(e)

    def __delattr__(self, key):
        super(ConnectionConfig, self).__delattr__(key)
        try:
            super(ConnectionConfig, self).__delitem__(key)
        except KeyError as e:
            raise AttributeError(e)

    def __delitem__(self, key):
        super(ConnectionConfig, self).__delitem__(key)
        try:
            super(ConnectionConfig, self).__delattr__(key)
        except AttributeError as e:
            raise KeyError(e)



import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import binascii

try:
    import tensorflow as tf
except:
    print("Tensorflow is not installed!")

""" This is an Experimental API Library Subject to Change


Example useage for train test split

rand_max = 1
max_frequency_bin=num_cols
num_steps=row_size
min_noise=-10
max_noise=10
mask_value=127
batch_size =32

data  = tf.data.Dataset.from_tensor_slices((x_train, y_train))
noise_ds = data.map(lambda image, label: tf_add_image_noise(image, label, rand_max, min_noise, max_noise))
freq_ds = noise_ds.map(lambda image, label: tf_mask_image_row(image, label, rand_max, max_frequency_bin, mask_value))
masked_ds = freq_ds.map(lambda image, label: tf_mask_image_column(image, label, num_steps, rand_max, max_frequency_bin, mask_value))
shuffle_aug = masked_ds.shuffle(buffer_size=x_train.shape[0], reshuffle_each_iteration=True).batch(batch_size)
shuffle_ds = data.shuffle(buffer_size=x_train.shape[0], reshuffle_each_iteration=True).batch(batch_size)

 """


def convert_tf_lite(tflite_model):
    return binascii.hexlify(tflite_model).decode("ascii")


def visualize_features(
    event_index, x_train, y_train, row_size, num_cols, class_map=None, figsize=(16, 8)
):
    if class_map:
        reverse_class_map = {int(v): k for k, v in class_map.items()}
        print("Class: ", reverse_class_map[np.argmax(y_train[event_index])])
    else:
        print("Class: ", np.argmax(y_train[event_index]))
    fig, axs = plt.subplots(1, 2, figsize=figsize)
    sns.heatmap(x_train[event_index].reshape(row_size, num_cols).T, ax=axs[0])
    plt.plot(x_train[event_index].reshape(row_size, num_cols))
    plt.ylim((0, 256))
    plt.show()


def add_tensor_noise(image, rand_max=1, min_noise=-10, max_noise=10):
    if np.random.randint(0, rand_max):
        image_numpy = image.numpy()
        image = tf.constant(
            image_numpy + np.random.randint(min_noise, max_noise, size=image.shape),
            dtype=image.dtype,
            shape=image.shape,
            name="Const",
        )

    return image


def visualize_tf_data(data, index, row_size, num_cols):
    count = 0
    for image, label in data.take(index):
        count += 1
        if count == index:
            fig = plt.figure(figsize=(row_size + 2, 4))
            sns.heatmap(image.numpy().reshape(-1, num_cols).T)
            plt.title("Label: {}".format(np.argmax(label)))
            plt.show()


def mask_tensor_row(image, rand_max=1, max_frequency_bin=1, mask_value=127):
    for _ in range(np.random.randint(0, rand_max)):
        start = np.random.randint(0, max_frequency_bin)
        image_numpy = image.numpy()
        np.put(
            image_numpy, np.arange(start, image.shape[0], max_frequency_bin), mask_value
        )
        image = tf.constant(
            image_numpy, dtype=image.dtype, shape=image.shape, name="Const"
        )

    return image


def mask_tensor_columns(
    image, num_cols, rand_max=1, max_frequency_bin=1, mask_value=127
):
    num_masks = np.random.randint(0, rand_max)
    for _ in range(num_masks):
        start = np.random.randint(0, num_cols)
        image_numpy = image.numpy()
        np.put(
            image_numpy,
            np.arange(
                start * max_frequency_bin, start * max_frequency_bin + max_frequency_bin
            ),
            mask_value,
        )
        image = tf.constant(
            image_numpy, dtype=image.dtype, shape=image.shape, name="Const"
        )

    return image


def add_image_noise(
    image, label, rand_max=1, min_noise=10, max_noise=10, tf_int_type=tf.int32
):
    im_shape = image.shape
    [image] = tf.py_function(
        add_tensor_noise, [image, rand_max, min_noise, max_noise], [tf_int_type]
    )
    image.set_shape(im_shape)

    return image, label


def mask_image_row(
    image, label, rand_max=1, max_frequency_bin=1, mask_value=127, tf_int_type=tf.int32
):
    im_shape = image.shape
    [image] = tf.py_function(
        mask_tensor_row, [image, rand_max, max_frequency_bin, mask_value], [tf_int_type]
    )
    image.set_shape(im_shape)

    return image, label


def mask_image_column(
    image,
    label,
    num_cols,
    rand_max=1,
    max_frequency_bin=1,
    mask_value=127,
    tf_int_type=tf.int32,
):
    im_shape = image.shape
    [image] = tf.py_function(
        mask_tensor_columns,
        [image, num_cols, rand_max, max_frequency_bin, mask_value],
        [tf_int_type],
    )
    image.set_shape(im_shape)

    return image, label


def plot_training_results(
    tf_model,
    history,
    x_train,
    y_train,
    x_validate,
    y_validate,
    class_map=None,
    figsize=(32, 10),
):
    fig, axs = plt.subplots(2, 2, figsize=figsize)
    epochs = range(1, len(history["loss"]) + 1)
    axs[0][0].plot(epochs, history["loss"], "g.", label="Training los    s")
    axs[0][0].plot(epochs, history["val_loss"], "b", label="Validation loss")
    axs[0][0].set_title("Training and validation loss")
    axs[0][0].set_xlabel("Epochs")
    axs[0][0].set_ylabel("Loss")
    axs[1][0].plot(epochs, history["accuracy"], "g.", label="Training Accuracy")
    axs[1][0].plot(epochs, history["val_accuracy"], "b", label="Validation Accuracy")
    axs[1][0].set_title("Training and validation Accuracy")
    axs[1][0].set_xlabel("Epochs")
    axs[1][0].set_ylabel("Accuracy")

    sns.heatmap(
        tf.math.confusion_matrix(
            np.argmax(tf_model.predict(x_train), axis=1), np.argmax(y_train, axis=1)
        ),
        annot=True,
        ax=axs[0][1],
    )

    axs[0][1].set_title("Training Confusion Matrix")

    sns.heatmap(
        tf.math.confusion_matrix(
            np.argmax(tf_model.predict(x_validate), axis=1),
            np.argmax(y_validate, axis=1),
        ),
        annot=True,
        ax=axs[1][1],
    )

    axs[1][1].set_title("Validation Confusion Matrix")

    plt.show()


def reset_weights(model):
    for layer in model.layers:
        if isinstance(layer, tf.keras.Model):  #
            reset_weights(layer)
            continue
        if hasattr(layer, "cell"):
            init_container = layer.cell
        else:
            init_container = layer

        for key, initializer in init_container.__dict__.items():
            if "initializer" not in key:
                continue

            if key == "recurrent_initializer":
                var = getattr(init_container, "recurrent_kernel")
            else:
                var = getattr(init_container, key.replace("_initializer", ""))

            var.assign(initializer(var.shape, var.dtype))
            # use the initializer



class DskException(Exception):
    """Base KB_DSK_Basic Exception"""

    pass


class QueryExistsException(DskException):
    pass


class BadArgumentException(DskException):
    pass


class BackendErrorException(DskException):
    pass


class FeatureFileExistsException(DskException):
    def __init__(self):
        Exception.__init__(
            self, "FeatureFile already exists. Use force=True to override."
        )


class PipelineOrderException(Exception):
    pass


class PipelineDataColumnsException(Exception):
    def __init__(self):
        msg = """Warning: Binary generation is currently only possible for the following sensors:
* ACCELEROMETERX
* ACCELEROMETERY
* ACCELEROMETERZ
* GYROSCOPEX
* GYROSCOPEY
* GYROSCOPEZ
If you are using these sensors but have different names, change the column
names of your data to align with the above to generate device code."""
        print(msg)


class DuplicateValueError(Exception):
    def __init__(self, message="Duplicate values used"):
        super(DuplicateValueError, self).__init__(message)



import os
import json
from sensiml.base import utility


class Function(object):
    """Base class for a transform object"""

    _uuid = ""
    _type = None
    _name = None
    _function_in_file = ""
    _description = ""
    _input_contract = ""
    _subtype = ""
    _has_c_version = False
    _library_pack = ""
    _automl_available = False

    def __init__(self, connection):
        self._connection = connection

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def type(self):
        return self._type

    @property
    def has_c_version(self):
        return self._has_c_version

    @has_c_version.setter
    def has_c_version(self, value):
        self._has_c_version = value

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def input_contract(self):
        return self._input_contract

    @input_contract.setter
    def input_contract(self, value):
        self._input_contract = value

    @property
    def subtype(self):
        return self._subtype

    @subtype.setter
    def subtype(self, value):
        self._subtype = value

    @property
    def library_pack(self):
        return self._library_pack

    @library_pack.setter
    def library_pack(self, value):
        self._library_pack = value

    @property
    def automl_available(self):
        return self._automl_available

    @automl_available.setter
    def automl_available(self, value):
        self._automl_available = value

    def get_function_info(self):
        """Returns function properties after checking integrity constraints.

        Returns:
            (dict): containing the function's name, type, subtype, and has_c_version

        Raises:
            error if function does not have a name
            error if function does not have a type
        """
        # Sanity Checks
        try:
            assert self.name is not None
        except Exception as e:
            print("Function name cannot be None")
            raise (e)

        try:
            assert self.type is not None
        except Exception as e:
            print(
                "Function type cannot be None, please set to the correct type (ie. Transform, Optimizer, etc.)"
            )
            raise (e)

        function_info = {
            "name": self.name,
            "type": self.type,
            "subtype": self.subtype,
            "has_c_version": self.has_c_version,
            "library_pack": self.library_pack,
            "automl_available": self.automl_available,
        }

        return function_info

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""
        url = "transform/{0}/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.name = response_data["name"]
            self.type = response_data["type"]
            self.subtype = response_data["subtype"]
            self.description = response_data["description"]
            self.input_contract = response_data["input_contract"]
            self.has_c_version = response_data["has_c_version"]
            self.library_pack = response_data["library_pack"]
            self.automl_available = response_data["automl_available"]

        return response_data

    def data(self):
        """Calls the REST API and request a transform file's binary data."""
        url = "transform/{0}/data/".format(self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response_data

    def initialize_from_dict(self, input_dictionary):
        """Populates a single transform object from a dictionary of properties from the server.

        Args:
            input_dictionary (dict): containing uuid, type, subtype, name, function_in_file, description,
            input_contract, and subtype
        """
        self.uuid = input_dictionary["uuid"]
        self.type = input_dictionary["type"]
        self.subtype = input_dictionary["subtype"]
        self.name = input_dictionary["name"]
        self.description = input_dictionary["description"]
        self.input_contract = input_dictionary["input_contract"]
        self.subtype = input_dictionary["subtype"]
        self.has_c_version = input_dictionary["has_c_version"]
        self.library_pack = input_dictionary["library_pack"]
        self.automl_available = input_dictionary["automl_available"]

    def __str__(self):
        input_string = ""
        for input_ in [
            i
            for i in self.input_contract
            if i["name"]
            not in [
                "classifiers",
                "validation_methods",
                "number_of_times",
                "sample_size",
            ]
        ]:
            if "element_type" in input_:
                input_string += "\n    {name} ({type} of {element_type})".format(
                    **input_
                )
            else:
                input_string += "\n    {name} ({type})".format(**input_)
            if "options" in input_:
                options_string = ""
                for option in input_["options"]:
                    options_string += "{name}, ".format(**option)
                input_string += " {{options: {0}}}".format(options_string[:-2])
            if "description" in input_:
                input_string += ": {description}".format(**input_)

        return (
            "NAME: {0}\n"
            "TYPE: {1}\n"
            "SUBTYPE: {2}\n"
            "DESCRIPTION: {3}\n"
            "LIBRARY PACK {4}\n"
        ).format(
            self.name, self.type, self.subtype, self.description, self.library_pack
        )



import logging

import sensiml.base.utility as utility
from sensiml.datamanager.base import Base, BaseSet
from sensiml.datamanager.label import Label
from sensiml.datamanager.labelvalue import LabelValue
from sensiml.datamanager.segmenter import Segmenter

logger = logging.getLogger(__name__)


class Segment(Base):
    """Base class for a label object."""

    _fields = [
        "uuid",
        "sample_start",
        "sample_end",
        "segmenter",
        "label",
        "label_value",
    ]

    _field_map = {
        "sample_start": "capture_sample_sequence_start",
        "sample_end": "capture_sample_sequence_end",
    }

    def __init__(
        self, connection, project, capture, segmenter=None, label=None, label_value=None
    ):
        """Initialize a metadata object.

        Args:
            connection
            project
            capture
        """
        super().__init__(connection)

        self._uuid = ""
        self._sample_start = 0
        self._sample_end = 0
        self._connection = connection
        self._project = project
        self._capture = capture
        self._segmenter = segmenter
        self._label = label
        self._label_value = label_value

    @property
    def uuid(self):
        """Auto generated unique identifier for the metadata object"""
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def sample_start(self):
        """The index of the first sample of the label"""
        return self._sample_start

    @sample_start.setter
    def sample_start(self, value):
        self._sample_start = value

    @property
    def sample_end(self):
        """The index of the last sample of the label"""
        return self._sample_end

    @sample_end.setter
    def sample_end(self, value):
        self._sample_end = value

    @property
    def segmenter(self):
        """The index of the last sample of the label"""
        if isinstance(self._segmenter, Segmenter):
            return self._segmenter.uuid

        return self._segmenter

    @segmenter.setter
    def segmenter(self, value):
        self._segmenter = value

    @property
    def label(self):
        if isinstance(self._label, Label):
            return self._label.uuid
        else:
            return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @property
    def label_value(self):
        if isinstance(self._label_value, LabelValue):
            return self._label_value.uuid
        else:
            return self._label_value

    @label_value.setter
    def label_value(self, value):
        self._label_value = value

    def insert(self):
        """Calls the REST API and inserts a metadata object onto the server using the local object's properties."""
        url = "project/{0}/capture/{1}/label-relationship/".format(
            self._project.uuid, self._capture.uuid
        )

        data = self._to_representation()

        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

        return response

    def update(self):
        """Calls the REST API and updates the object on the server."""
        self._capture.await_ready()
        url = "project/{0}/capture/{1}/label-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        data = self._to_representation()

        response = self._connection.request("put", url, data)
        response_data, err = utility.check_server_response(response)

        return response

    def delete(self):
        """Calls the REST API and deletes the object from the server."""
        url = "project/{0}/capture/{1}/label-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

        return response

    def refresh(self):
        """Calls the REST API and populates the local object's properties from the server."""
        url = "project/{0}/capture/{1}/label-relationship/{2}/".format(
            self._project.uuid, self._capture.uuid, self.uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.initialize_from_dict(response_data)

        return response


class SegmentSet(BaseSet):
    """Base class for a collection of segments"""

    def __init__(self, connection, project, capture, initialize_set=True):
        self._connection = connection
        self._project = project
        self._capture = capture
        self._set = None
        self._objclass = Segment

        if initialize_set:
            self.refresh()

    @property
    def get_set_url(self):
        return "project/{0}/capture/{1}/label-relationship/".format(
            self._project.uuid, self._capture.uuid
        )

    @property
    def segments(self):
        return self.objs

    def _new_obj_from_dict(self, data):
        """Creates a new label from data in the dictionary.

        Args:
            dict (dict): contains label properties uuid, name, type, value, capture_sample_sequence_start, and
            capture_sample_sequence_end

        Returns:
            label object

        """
        segment = self._objclass(self._connection, self._project, self._capture)
        segment.initialize_from_dict(data)
        return segment



class LoginError(Exception):
    pass


class MethodDeprecatedError(LoginError):
    def __init__(self, message="This method is no longer allowed.", *args):
        super(MethodDeprecatedError, self).__init__(message, *args)


class LoginAttemptsExceededError(LoginError):
    def __init__(self, message="Maximum number of login attempts exceeded.", *args):
        super(LoginAttemptsExceededError, self).__init__(message, *args)


class ConnectionAbortedError(LoginError):
    def __init__(self, message="The connection was aborted by the user.", *args):
        super(ConnectionAbortedError, self).__init__(message, *args)



from __future__ import print_function

import json
import logging
import time

import requests
from pandas import DataFrame
from six.moves import input

logger = logging.getLogger(__name__)


def prompt_for_overwrite(file_that_exists):

    acceptable = {"yes": True, "y": True, "n": False, "no": False}

    prompts = 0
    while prompts < 3:
        print("{0} Exists. Do you wish to overwrite?".format(file_that_exists))
        answer = input().lower()
        if answer in acceptable.keys():
            return acceptable[answer]
        else:
            print("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")
            prompts = prompts + 1
    return False


def data_response(response_json):
    """handle response with data from kbserver"""
    data = response_json.get("data", None)
    if data:
        if isinstance(data, list):
            print("Reason : {}".format(data[0].replace("DETAIL", "Details")))
        elif data:
            print("Reason : {}".format(data))


def detail_response(response_json, renderer=None):
    """handle responses with detail (typically a server trace) form kbserver"""
    detail = response_json.get("detail", None)
    if detail:
        if renderer:
            renderer.render(detail)
        print(detail)


def category_response_error(reason_json, status, renderer=None):
    """handle errors with category from kbserver"""
    msg = (
        "HTTPError {} : {}\n".format(status, reason_json.get("category", None))
        + "Reason : {}\n".format(reason_json.get("err", None))
        + "Details : {}".format(reason_json.get("extra", None))
    )

    if renderer:
        renderer.render(msg)

    print(msg)


def read_response_json_data(response, renderer=None):
    """The goal here is to give nice human readable errors to our users"""
    try:
        # If there is a json response we read it here
        response_json = response.json()
        if renderer:
            renderer.render("HTTPError: {}".format(response.status_code))
        print("\n\nHTTPError: {}".format(response.status_code))

        data_response(response_json)
        detail_response(response_json, renderer)
    except:
        pass

    try:
        # if there is a json in the reason we read it here
        category_response_error(
            json.loads(response.reason), response.status_code, renderer
        )
    except:
        pass


def check_server_response(response, is_octet=False, renderer=None):
    """Check server response messages to see if they're JSON or not. If JSON,
    check whether or not an error or errors were returned.
    """
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        read_response_json_data(response, renderer)
        raise

    if is_octet:
        return response, False

    try:
        data = response.json()
    except (ValueError, KeyError) as e:
        # Response was not JSON. Return content for now.
        return response.content, True

    return data, False


def check_pipeline_status(response_data, parse_results, silent=False, **kwargs):
    """Read the response for a pipeline sandbox for kbcloud

    Args:
        response_data (dictionary): rthe body of the response from the server
        parse_results (function): a function that will parse the result upon success
        silent (bool, False): silence the Pending and Started messages
        **kwargs : contains arguments to pass through to parse_results function
    """
    # If there are server errors, show them
    if response_data.get("errors") and len(response_data["errors"]):
        for error in response_data["errors"]:
            if error.get("detail"):
                logger.warning("{error}: {message} ({detail})".format(**error))
            else:
                logger.warning("{error}: {message}".format(**error))

    # If there is a status, the pipeline is Pending, Not Running, Failed, or in the Queue
    status = response_data.get("status", "SUCCESS")
    if status in ["PENDING", "STARTED", "SENT"]:
        if not silent:
            print(response_data.get("message", None), end="")
        return None, response_data.get("message", None)
    elif status in ["FAILURE", "REVOKED", None]:
        print(response_data.get("message", None))
        return response_data, None

    return parse_results(response_data, **kwargs)


def wait_for_pipeline_result(
    requester,
    lock=True,
    silent=True,
    wait_time=15,
    skip_printing=4,
    renderer=None,
    **kwargs
):
    count = 0
    start = time.time()

    results = requester.retrieve(silent=silent, **kwargs)
    if results[0] is not None:
        lock = False

    while lock:
        time.sleep(wait_time)
        results = requester.retrieve(silent=True, **kwargs)
        if results[0] is None:
            if renderer:
                renderer.render(results[1])
            if count % skip_printing == 0:
                print("")
                print(results[1], end="")
                print(" ", end="")
            else:
                print(".", end="")
            count += 1
        else:
            lock = False

    execution_time = time.time() - start
    msg = "\n\nResults Retrieved... Execution Time: {0} min. {1} sec.".format(
        int(execution_time / 60), int(execution_time % 60)
    )

    if renderer:
        renderer.render(msg)

    print(msg)

    return results


def make_statistics_table(response_data):
    error_report = None
    data = DataFrame(response_data)

    if "event_labels" in data.columns:
        labels_dict = {}
        labels = json.loads(data["event_labels"].to_json(orient="records"))
        data = data.drop("event_labels", 1)
        for label in labels:
            for name, info in label.items():
                for values in info:
                    column_name = "{0}: {1}".format(name, values["value"])
                    if column_name not in labels_dict:
                        labels_dict[column_name] = []
                    labels_dict[column_name].append(values["count"])
        for column_name, column_values in labels_dict.items():
            data[column_name] = column_values

    if "metadata" in data.columns:
        meatadata_dict = {}
        meatadata_set = json.loads(data["metadata"].to_json(orient="records"))
        data = data.drop("metadata", 1)
        for metadata in meatadata_set:
            for name, value in metadata.items():
                if name not in meatadata_dict:
                    meatadata_dict[name] = []
                meatadata_dict[name].append(value)
        for column_name, column_values in meatadata_dict.items():
            data[column_name] = column_values

    # Post-processing to get numeric index
    data["sample"] = data.index.astype(int)
    data = data.set_index("sample")
    data.index.names = [None]
    data = data.sort_index(axis=0)
    cols = data.columns.tolist()
    if "id" in cols:
        data = data[["id"] + [c for c in cols if c != "id"]]

    return data.sort_index(axis=0), error_report


def make_cost_report(device_config, cost_dict=None, neuron_array=None):
    """A printed tabular report of the device costs incurred by the KnowledgePack or Sandbox"""
    cost_columns = ["Bytes", "RAM", "Microseconds", "Stack"]
    cost_columns_display_names = ["Flash", "SRAM", "Latency", "Stack"]
    cost_units = ["(Bytes)", "(Bytes)", "(Microseconds)", "(Bytes)"]
    na_string = "0"
    not_specified_string = "Not specified"

    number_of_costs = len(cost_columns)

    def horizontal_line():
        print("-" * (45 + 17 * number_of_costs))

    def val_or_string(dict, key, string):
        if key in dict and dict[key]:
            if 0 < float(dict[key]) < 1:
                return round(float(dict[key]), 1)
            else:
                return int(float(dict[key]))
        else:
            return string

    def sum_or_string(summary, key, string):
        items_to_sum = [x[key] for x in summary["pipeline"] if key in x and x[key]]
        for a in ["neurons", "sensors"]:
            if summary.get(a, False) and summary[a].get(key, False):
                items_to_sum.append(summary[a][key])
        return int(sum(items_to_sum)) if len(items_to_sum) else string

    def populate_cost_keys(columns, cost_dict, string):
        all_costs = {cost: string for cost in columns}
        all_costs.update(cost_dict)
        return all_costs

    try:
        budget = device_config["budget"]
    except:
        budget = None

    horizontal_line()
    format_columns = "{:45}" + "{:>17}" * number_of_costs
    print(format_columns.format("", *cost_columns_display_names))
    print(format_columns.format("", *cost_units))
    horizontal_line()

    format_budget = "{:45}"
    for cost in cost_columns:
        format_budget += "{" + cost + ":>17}"

    if budget:
        full_budget = populate_cost_keys(cost_columns, budget, not_specified_string)
        print(
            format_budget.format(
                "Budgets ("
                + device_config["target_platform"]
                + " "
                + device_config["platform_version"]
                + ")",
                **full_budget
            )
        )
    else:
        print("There is no budget defined")

    if cost_dict.get("neurons", None):
        print("\n\nPattern Matching Engine Costs")
        horizontal_line()
        array_costs = populate_cost_keys(cost_columns, cost_dict["neurons"], na_string)
        array_size = "Classifier 1: {} neurons x {} features".format(
            len(neuron_array), len(neuron_array[0]["Vector"])
        )
        print(format_budget.format(array_size, **array_costs))
        horizontal_line()

    print("\n\nFeature Extraction Costs")
    horizontal_line()
    step_format = "{:45}" + "{:>17}" * number_of_costs
    step_format_sub = "{:6}{:39}" + "{:>17}" * number_of_costs
    fe_subtotals = [0] * number_of_costs
    for i, step in enumerate(cost_dict["pipeline"]):
        costs = [val_or_string(step, cost, na_string) for cost in cost_columns]
        fe_subtotals = [
            a + (int(b) if b == "0" else b) for a, b in zip(fe_subtotals, costs)
        ]
        if step["type"] in ["transform", "segmenter"]:
            print(
                step_format.format(
                    " - {} ({})".format(step["name"], step["type"]), *costs
                )
            )
        elif step["type"] == "generatorset":
            print(
                step_format.format(
                    " - {} ({})".format(step["name"], step["type"]),
                    *["" for cost in cost_columns]
                )
            )
            if step.get("per_generator_costs", None):
                for feature_generator in step["per_generator_costs"].keys():
                    print(
                        step_format_sub.format(
                            "",
                            "{} x {}".format(
                                step["per_generator_costs"][feature_generator][
                                    "num_iterations"
                                ],
                                feature_generator,
                            ),
                            *[
                                val_or_string(
                                    step["per_generator_costs"][feature_generator],
                                    cost,
                                    na_string,
                                )
                                for cost in cost_columns
                            ]
                        )
                    )
    horizontal_line()

    format_fe_totals = "{:>45}" + "{:>17}" * number_of_costs
    print(format_fe_totals.format("Feature Extraction Subtotals:", *fe_subtotals))

    if cost_dict.get("framework", None):
        print("\n\nFramework Costs")
        horizontal_line()
        framework_costs = populate_cost_keys(
            cost_columns, cost_dict["framework"], na_string
        )
        framework_size = "Knowledge Pack Overhead"
        print(format_budget.format(framework_size, **framework_costs))

    horizontal_line()

    totals = {cost: sum_or_string(cost_dict, cost, na_string) for cost in cost_columns}
    cost_dict_minus_sensors = {k: v for k, v in cost_dict.items() if k != "sensors"}
    totals_minus_sensors = {
        cost: sum_or_string(cost_dict_minus_sensors, cost, na_string)
        for cost in cost_columns
    }
    format_totals = "{:>45}" + "{:>17}" * number_of_costs
    print(
        format_totals.format(
            "Totals Excluding Sensor Costs:",
            *[totals_minus_sensors[cost] for cost in cost_columns]
        )
    )

    if cost_dict.get("sensors", None):
        print("\n\nSensor Costs")
        horizontal_line()
        sensor_costs = populate_cost_keys(cost_columns, cost_dict["sensors"], na_string)
        sensor_size = "Buffer Size: {} samples x {} sensor streams".format(
            cost_dict["sensors"]["max_segment_length"],
            cost_dict["sensors"]["number_of_sensors"],
        )
        print(format_budget.format(sensor_size, **sensor_costs))

    horizontal_line()

    print(
        format_totals.format(
            "Totals Including Sensor Costs:", *[totals[cost] for cost in cost_columns]
        )
    )

    print("\nWarnings:")
    horizontal_line()
    over_budget = False
    if budget:
        full_budget = {cost: na_string for cost in cost_columns}
        full_budget.update(budget)
        for key in full_budget.keys():
            if totals[key] > full_budget[key] and full_budget[key] not in [
                na_string,
                not_specified_string,
            ]:
                print(
                    "   {} {} exceeds budget for {} {} ({} available)".format(
                        totals[key],
                        key,
                        device_config["target_platform"],
                        device_config["platform_version"],
                        full_budget[key],
                    )
                )
                over_budget = True
    if not over_budget and cost_dict.get("neurons", None):
        print("   None")
    if cost_dict.get("neurons", None) is None:
        print(
            "   This table does not include the costs of the neuron array or sensor buffer"
        )
        print("   For a more complete picture, see the cost report of a Knowledge Pack")
    print("\n")



from sensiml.method_calls.augmentationcall import AugmentationCall
from sensiml.method_calls.augmentationcallset import AugmentationCallSet
from sensiml.method_calls.capturefilecall import CaptureFileCall
from sensiml.method_calls.classifiercall import ClassifierCall
from sensiml.method_calls.datafilecall import DataFileCall
from sensiml.method_calls.featurefilecall import FeatureFileCall
from sensiml.method_calls.functioncall import FunctionCall
from sensiml.method_calls.generatorcall import GeneratorCall
from sensiml.method_calls.generatorcallset import GeneratorCallSet
from sensiml.method_calls.optimizercall import OptimizerCall
from sensiml.method_calls.querycall import QueryCall
from sensiml.method_calls.selectorcall import SelectorCall
from sensiml.method_calls.selectorcallset import SelectorCallSet
from sensiml.method_calls.trainandvalidationcall import TrainAndValidationCall
from sensiml.method_calls.trainingalgorithmcall import TrainingAlgorithmCall
from sensiml.method_calls.validationmethodcall import ValidationMethodCall

__all__ = [
    "FunctionCall",
    "QueryCall",
    "FeatureFileCall",
    "TrainingAlgorithmCall",
    "AugmentationCall",
    "AugmentationCallSet",
    "GeneratorCall",
    "GeneratorCallSet",
    "SelectorCall",
    "SelectorCallSet",
    "ValidationMethodCall",
    "ClassifierCall",
    "OptimizerCall",
    "TrainingAlgorithmCall",
    "CaptureFileCall",
    "DataFileCall",
]



import ctypes
import json
import os
import subprocess
import zipfile
from ctypes import CDLL
from tkinter import E

try:
    from typing import Literal
except:
    from typing_extensions import Literal

try:
    from wurlitzer import pipes
except:
    pass

from typing import Dict, List, Optional

from numpy import ndarray
from pandas import DataFrame, Series

USE_PIPES = True if os.name == "posix" else False


class ImportPathException(Exception):
    pass


def get_activation(tensor):
    return (max(tensor) + 128) / 256.0


def dummy_function(*args, **kwrags):
    print("Not Supported by this version of the Knowledge Pack.")
    return None


def compile_libsensiml_shared_library(path):

    current_dir = os.getcwd()
    try:
        os.chdir(path)

        if os.path.exists("libtensorflow-microlite.a"):
            print("calling: g++ --shared libtensorflow-microlite.a")
            subprocess.call(
                "g++ --shared -o libsensiml.so -Wl,--whole-archive libtensorflow-microlite.a libsensiml.a  -Wl,--no-whole-archive",
                shell=True,
            )
        else:
            print("calling: ar -x libsensiml.a")
            subprocess.call(["ar", "-x", "libsensiml.a"])
            print("calling: g++ -shared -o libsensiml.so *.o")
            subprocess.call("g++ -shared -o libsensiml.so *.o", shell=True)

        if os.path.exists("libsensiml.so"):
            print("Successfully created libsensiml.so")
    except Exception as e:
        raise e
    finally:
        os.chdir(current_dir)

    os.chdir(current_dir)


def parse_model_json_for_class_map(path):

    if os.path.exists(os.path.join(path, "model.json")):
        model_info = json.load(open(os.path.join(path, "model.json"), "r"))
        class_maps = [x["ClassMaps"] for x in model_info["ModelDescriptions"]]
        return [{int(k): v for k, v in class_map.items()} for class_map in class_maps]

    return None


def parse_model_json_for_model_type(path):

    if os.path.exists(os.path.join(path, "model.json")):
        model_info = json.load(open(os.path.join(path, "model.json"), "r"))
        return [
            description["ModelType"] for description in model_info["ModelDescriptions"]
        ]

    return None


def is_tensorflow(ModelType):

    if type(ModelType) != str:
        return False

    if "TF Micro" in ModelType:
        return True

    if "TensorFlow Lite for Microcontrollers" in ModelType:
        return True

    return False


class struct_pme_pattern(ctypes.Structure):
    __slots__ = ["influence", "category", "vector"]

    _fields_ = [
        ("influence", ctypes.c_uint16),
        ("category", ctypes.c_uint16),
        ("vector", ctypes.POINTER(ctypes.c_uint8)),
    ]


class struct_pme_model_header(ctypes.Structure):
    __slots__ = ["number_patterns", "pattern_length"]

    _fields_ = [
        ("number_patterns", ctypes.c_uint16),
        ("pattern_length", ctypes.c_uint16),
    ]


class struct_model_result(ctypes.Structure):
    __slots__ = ["num_outputs", "output_tensor"]

    _fields_ = [
        ("num_outputs", ctypes.c_ubyte),
        ("output_tensor", ctypes.POINTER(ctypes.c_int16)),
    ]


def empty_function(*args, **kwargs):
    print("Not Supported by this Knowledge Pack.")
    return


class Model(object):
    def __init__(self, neuron_array, class_map, configuration, feature_summary=None):
        self._class_map = None
        self._feature_summary = None
        self._neuron_array = None
        self._configuration = None

        self.class_map = class_map
        self.neuron_array = neuron_array
        self.configuration = configuration

        if feature_summary:
            self.feature_summary = feature_summary
        else:
            self.feature_summary_from_neurons()

    @property
    def class_map(self):
        return self._class_map

    @class_map.setter
    def class_map(self, value):
        self._class_map = value

    @property
    def neuron_array(self):
        return self._neuron_array

    @neuron_array.setter
    def neuron_array(self, value):
        self._neuron_array = value

    @property
    def configuration(self):
        return self._configuration

    @configuration.setter
    def configuration(self, value):
        self._configuration = value

    @property
    def feature_summary(self):
        return self._feature_summary

    @feature_summary.setter
    def feature_summary(self, value):
        self._feature_summary = [{"Feature": str(x)} for x in value]

    def feature_summary_from_neurons(self):
        feature_vector = self.neuron_array[0]["Vector"]
        self.feature_summary = [{"Feature": str(x)} for x in range(len(feature_vector))]


class SMLRunner:
    def __init__(self, path: str, class_map: Optional[Dict] = None):
        """
        This class provides a Python interface to the Knowledge Pack library C code.

        Args:
            path (str): Path to the downloaded Knowledge Pack .zip, the libsensiml.so, or libsensiml.dll.
            class_map (Dict, optional): The class map to use for this model. Defaults to None.

        Example:

            After downloading the Knowledge Pack for Windows or Linux:

                sml = SMLRunner(<path_to_knowledgepack.zip>)
                sml.init_model()

        """

        self._run_type = None
        self._model_initialized = False

        if path[-3:] == "zip":
            unzip_dir = os.path.basename(path)[:-4]
            with zipfile.ZipFile(path, "r") as zip_ref:
                zip_ref.extractall(unzip_dir)
                lib_path = os.path.join(unzip_dir, "knowledgepack", "sensiml", "lib")
                model_json_path = os.path.join(unzip_dir, "knowledgepack")

        if os.name == "nt":
            if not os.path.exists(os.path.join(lib_path, "libsensiml.dll")):
                print("ERROR: libsensiml.dll as not found in {}.".format(lib_path))
                raise ImportPathException(
                    "ERROR: libsensiml.dll as not found in {}".format(lib_path)
                )

            clf_lib = CDLL(os.path.join(lib_path, "libsensiml.dll"))

        if os.name == "posix":
            if not os.path.exists(os.path.join(lib_path, "libsensiml.so")):
                print(
                    "Warning: libsensiml.so as not found in {}. Attempting".format(
                        lib_path
                    )
                )
                if not os.path.exists(os.path.join(path, "libsensiml.a")):
                    print("ERROR: libsensiml.a as not found in {}.".format(lib_path))

                print("Attempting to compile libsensiml.so")

                compile_libsensiml_shared_library(lib_path)

                if not os.path.exists(os.path.join(lib_path, "libsensiml.so")):
                    raise ImportPathException(
                        "ERROR: libsensiml.so as not found in {}".format(lib_path)
                    )

            clf_lib = CDLL(os.path.join(lib_path, "libsensiml.so"))

        if class_map is not None:
            if isinstance(class_map, dict):
                self.class_map = [{int(k): v for k, v in class_map.items()}]
            elif isinstance(class_map, list):
                self.class_map = []
                for model_class_map in class_map:
                    self.class_map.append(
                        {int(k): v for k, v in model_class_map.items()}
                    )
        else:
            self.class_map = parse_model_json_for_class_map(model_json_path)

        self.model_type = parse_model_json_for_model_type(model_json_path)

        self._model_init = clf_lib.kb_model_init
        self._model_init.argtypes = []

        self._run_model = clf_lib.kb_run_model
        self._run_model.argtypes = [
            ctypes.POINTER(ctypes.c_int16),
            ctypes.c_int,
            ctypes.c_int,
        ]
        self._run_model.restype = ctypes.c_int

        self._run_model_cascade_features = clf_lib.kb_run_model_with_cascade_features
        self._run_model_cascade_features.argtypes = [
            ctypes.POINTER(ctypes.c_int16),
            ctypes.c_int,
            ctypes.c_int,
        ]
        self._run_model_cascade_features.restype = ctypes.c_int

        self._run_segment = clf_lib.kb_run_segment
        self._run_segment.argtypes = [ctypes.c_int]
        self._run_segment.restype = ctypes.c_int

        self._add_segment = clf_lib.kb_add_segment
        self._add_segment.argtypes = [
            ctypes.POINTER(ctypes.c_int16),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
        ]

        self._reset_model = clf_lib.kb_reset_model
        self._reset_model.argtypes = [ctypes.c_int]
        self._reset_model.restype = ctypes.c_int

        self._kb_get_model_sg_length = clf_lib.kb_get_model_sg_length
        self._kb_get_model_sg_length.argtypes = [
            ctypes.c_int,
        ]
        self._kb_get_model_sg_length.restype = ctypes.c_int

        self._kb_get_sg_start_index = clf_lib.kb_get_sg_start_index
        self._kb_get_sg_start_index.argtypes = [
            ctypes.c_int,
        ]
        self._kb_get_sg_start_index.restype = ctypes.c_int

        self._sml_get_segment_length = clf_lib.sml_get_segment_length
        self._sml_get_segment_length.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int),
        ]

        self._sml_get_feature_bank_number = clf_lib.sml_get_feature_bank_number
        self._sml_get_feature_bank_number.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int),
        ]

        self._kb_get_segment_data = clf_lib.kb_get_segment_data
        self._kb_get_segment_data.argtypes = [
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int16),
        ]

        self._kb_get_model_framelen = clf_lib.kb_get_model_framelen
        self._kb_get_model_framelen.argtypes = [
            ctypes.c_int,
        ]

        # self._print_model_map = clf_lib.kb_print_model_map

        # self._print_model_result = clf_lib.kb_print_model_result
        # self._print_model_result.argtypes = [ctypes.c_int, ctypes.c_int]

        # self._print_model_score = clf_lib.kb_print_model_score
        # self._print_model_score.argtypes = [ctypes.c_int]
        # self._print_model_score.restype = ctypes.c_int

        # self._score_model = clf_lib.kb_score_model
        # self._score_model.argtypes = [ctypes.c_int, ctypes.c_uint16]
        # self._score_model.restype = ctypes.c_int

        # self._retrain_model = clf_lib.kb_retrain_model
        # self._retrain_model.argtypes = [ctypes.c_int]
        # self._retrain_model.restype = ctypes.c_int

        if hasattr(clf_lib, "kb_flush_model_buffer"):
            self._flush_model_buffer = clf_lib.kb_flush_model_buffer
            self._flush_model_buffer.argtypes = [ctypes.c_int]
            self._flush_model_buffer.restype = ctypes.c_int
        else:
            self._flush_model_buffer = dummy_function

        self._add_last_pattern_to_model = clf_lib.kb_add_last_pattern_to_model
        self._add_last_pattern_to_model.argtypes = [
            ctypes.c_int,
            ctypes.c_uint16,
            ctypes.c_uint16,
        ]
        self._add_last_pattern_to_model.restype = ctypes.c_int

        self._kb_get_classification_result_info = (
            clf_lib.kb_get_classification_result_info
        )
        self._kb_get_classification_result_info.argtypes = [
            ctypes.c_int,
            ctypes.c_void_p,
        ]
        self._kb_get_classification_result_info.restype = ctypes.c_int

        self._add_custom_pattern_to_model = clf_lib.kb_add_custom_pattern_to_model
        self._add_custom_pattern_to_model.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_uint16,
            ctypes.c_uint16,
        ]
        self._add_custom_pattern_to_model.restype = ctypes.c_int

        # self._print_model_class_map = clf_lib.kb_print_model_class_map
        # self._print_model_class_map.argtypes = [ctypes.c_int, ctypes.c_char_p]

        self._get_model_header = clf_lib.kb_get_model_header
        self._get_model_header.argtypes = [
            ctypes.c_int,
            ctypes.POINTER(struct_pme_model_header),
        ]
        self._get_model_header.restype = ctypes.c_int

        self._get_model_pattern = clf_lib.kb_get_model_pattern
        self._get_model_pattern.argtypes = [
            ctypes.c_int,
            ctypes.c_int,
            ctypes.POINTER(struct_pme_pattern),
        ]
        self._get_model_pattern.restype = ctypes.c_int

        self._flush_model = clf_lib.kb_flush_model_buffer
        self._flush_model.argtypes = [ctypes.c_int]
        self._flush_model.restype = ctypes.c_int

        try:
            self._sml_recognition_run = clf_lib.sml_recognition_run
            self._sml_recognition_run.argtypes = [
                ctypes.POINTER(ctypes.c_int16),
                ctypes.c_int,
            ]
            self._sml_recognition_run.restype = ctypes.c_int
        except:
            self._sml_recognition_run = None

        try:
            self._get_feature_vector = clf_lib.kb_get_feature_vector
            self._get_feature_vector.argtypes = [
                ctypes.c_int,
                ctypes.POINTER(ctypes.c_uint8),
                ctypes.POINTER(ctypes.c_uint8),
            ]
        except:
            self._get_feature_vector = empty_function

        try:
            self._get_feature_vector = clf_lib.kb_get_feature_vector
            self._get_feature_vector.argtypes = [
                ctypes.c_int,
                ctypes.POINTER(ctypes.c_uint8),
                ctypes.POINTER(ctypes.c_uint8),
            ]
        except:
            self._get_feature_vector = empty_function

        try:
            self._set_feature_vector = clf_lib.kb_set_feature_vector
            self._set_feature_vector.argtypes = [
                ctypes.c_int,
                ctypes.POINTER(ctypes.c_uint8),
            ]
        except:
            self._set_feature_vector = empty_function

        try:
            self._set_feature_vector = clf_lib.kb_set_feature_vector
            self._set_feature_vector.argtypes = [
                ctypes.c_int,
                ctypes.POINTER(ctypes.c_uint8),
            ]
            self._set_feature_vector.restype = ctypes.c_int
        except:
            self._set_feature_vector = empty_function

        try:
            self._recognize_feature_vector = clf_lib.kb_recognize_feature_vector
            self._recognize_feature_vector.argtypes = [ctypes.c_int]
            self._recognize_feature_vector.restype = ctypes.c_uint16
        except:
            self._recognize_feature_vector = empty_function

    def _initialized(self):
        if not self._model_initialized:
            print("Initialize the model before running this function")
            return None

        return True

    def _run_with(self, run_type):
        if self._run_type is None:
            self._run_type = run_type

        if self._run_type != run_type:
            print(
                "This model has already been run with {}. You will need to restart your kernel to run in this mode.".format(
                    self._run_type
                )
            )
            return False

        return True

    def init_model(self):
        """
        This will initialize the parameters of all the models, it should be run once before
        running any other APIs.
        """
        if not self._model_initialized:
            self._model_initialized = True
            self._model_init()

    def flush_model_buffer(self, model_index):
        """
        This will flush the data buffer of a Knowledge Pack.
        """
        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        self._flush_model_buffer(model_index_ctype)

    def add_custom_pattern_to_model(self, model_index, feature_vector, category, aif):
        """
        Updates a PME model by adding the last pattern to the internal database
        with the specified category and influence field

        Args:
           model_index(int): Index of the model to update
           feature_vector(list): List of uint8 values that will be added to the database
           category(uint16): category of the vector to add
           aif(uint16): weight function for the new pattern

         Returns:
             int: 0 if model does not support dynamic updates, -1 if model cannot be updated anymore, 1 if model was successfully updated
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        category_ctype = ctypes.c_uint16(category)
        aif_ctype = ctypes.c_uint16(aif)

        feature_vector_array = (ctypes.c_uint8 * len(feature_vector))()

        for index, value in enumerate(feature_vector):
            feature_vector_array[index] = ctypes.c_uint8(value)

        return self._add_custom_pattern_to_model(
            model_index_ctype, feature_vector_array, category_ctype, aif_ctype
        )

    def add_last_pattern_to_model(self, model_index, category, aif):
        """
        Updates a PME model by adding the last feature vector created to the database
        with a new category and influence field

        Args:
           model_index(int): Index of the model to update
           category(uint16): category of the vector to add
           aif(uint16): weight function for the new pattern

        Returns:
            int: 0 if model does not support dynamic updates, -1 if model cannot be updated anymore, 1 if model was successfully updated
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        category_ctype = ctypes.c_uint16(category)
        aif_ctype = ctypes.c_uint16(aif)

        return self._add_last_pattern_to_model(
            model_index_ctype, category_ctype, aif_ctype
        )

    def score_model(self, model_index, category, silent=False):
        """
        Given the ground truth of the current feature vector, this will score the model
        internally

        Args:
           model_index(int): Index of the model to update
           category(uint16): category of the vector to add

        Returns:
           int: 0 if model does not support dynamic scoring, -1 if model cannot be scored, 1 if model was successfully scored
        """
        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        category_ctype = ctypes.c_uint16(category)

        res = None
        with pipes() as (out, _):
            res = self._score_model(model_index_ctype, category_ctype)

        if silent:
            print(out.read())

        return res

    def retrain_model(self, model_index: int = 0):
        """
        After a model has been scored, this will retrain the model according
        to the models online learning approach

        Args:
           model_index(int): Index of the model to update

        Returns:
           int: 0 if model does not support dynamic scoring, 1 if model was scored
        """
        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        ret = self._retrain_model(model_index_ctype)

        if ret == 0:
            print("Retraining is not supported/enabled for this model.")

        return ret

    def run_model(
        self,
        data_sample,
        model_index: int = 0,
        debug_log: bool = True,
        model_api: Literal[
            "run_model", "run_model_cascade_features", "sml_recognition_run"
        ] = "run_model",
    ):
        """
        This is the main entry point into the pipeline. It takes a single timepoint of data as an array. Adds that sample to the internal ring buffer
        checks for a segment, generates features if there is a segment,
        produces a classification and returns the result.

        Args:
            data_sample(array):  single timepoint of data as an array from all sensors
            nsensors(int): unused
            model_index(int): Index of the model to run
            debug_log(bool): Prints the debug output from the Knowledge Pack
            model_api(str): use run_model, run_model_cascade_features or sml_recognition_run. These correspond to the API's in the kb.h file

        Returns:
            Classification results will be 0 if unknown through the classification numbers
                you have. This function returns -1 when a segment hasn't yet been identified.

        """
        if not self._initialized():
            return

        if self._run_with("run_model") is False:
            return

        if isinstance(data_sample, DataFrame) or isinstance(data_sample, Series):
            data = data_sample.values
        elif isinstance(data_sample, ndarray):
            data = data_sample
        else:
            print("Input data Must be either dataframe or array.")
            return

        data_array = (ctypes.c_int16 * data.shape[0])()
        nsensors_ctype = ctypes.c_int(0)
        model_index_ctype = ctypes.c_int(model_index)

        for index, value in enumerate(data):
            data_array[index] = ctypes.c_int16(value)

        if model_api == "run_model":
            ret = self._run_model(data_array, nsensors_ctype, model_index_ctype)
        elif model_api == "sml_recognition_run":
            ret = self._sml_recognition_run(data_array, model_index_ctype)
        elif model_api == "run_model_cascade_features":
            ret = self._run_model_cascade_features(
                data_array, nsensors_ctype, model_index_ctype
            )
        else:
            raise Exception("Invalid model_api")

        return ret

    def recognize_capture(
        self,
        data: DataFrame,
        model_index: int = 0,
        model_api: Literal[
            "run_model", "run_model_cascade_features", "sml_recognition_run"
        ] = "run_model",
        get_feature_vector: bool = False,
        get_sensor_data: bool = False,
        get_output_tensor: bool = True,
        verbose: bool = True,
    ) -> DataFrame:
        """Run the model against a DataFrame of sensor data

        Args:
            data (DataFrame): Input sensor data
            model_index (int, optional): The index of model to use. Defaults to 0.
            model_api (Literal[&quot;run_model&quot;,&quot;run_model_cascade_features&quot;,&quot;sml_recognition_run&quot;], optional): Model API to use. Defaults to "run_model".
            get_feature_vector (bool, optional): Includes the feature vector output in the returned result. Defaults to False.
            get_sensor_data (bool, optional): Includes the sensor data in the returned result. Defaults to False.
            get_output_tensor (bool, optional): Includes the output tensor probabilities in the returned result (if applicable). Defaults to True.

        Returns:
            DataFrame: results of the model along with the feature_vectors and output probabilities
        """

        M = []

        status = [len(data) // 4, len(data) // 2, 3 * len(data) // 4]
        self.flush_model_buffer(model_index=model_index)

        for i in range(0, len(data)):
            # pass a single sample
            ret = self.run_model(
                data.iloc[i], model_index=model_index, model_api=model_api
            )

            if ret >= 0:

                output_columns = {
                    "label_value": ret
                    if self.get_class_map(model_index) is None
                    else self.get_class_map(model_index).get(ret, ret),
                    "capture_sample_sequence_start": i
                    + 1
                    - self.get_segment_length(model_index=model_index),
                    "capture_sample_sequence_end": i,
                }

                if get_feature_vector:
                    output_columns["feature_vector"] = self.get_feature_vector(
                        model_index
                    )

                n_sensors = self.get_model_framelen(model_index)

                if get_sensor_data:
                    for i in range(n_sensors):
                        sensor_data = self.get_segment_data(model_index)
                        output_columns["sensor_data_%d" % i] = sensor_data[i]

                if get_output_tensor:
                    output_tensor = self.get_model_result(model_index)
                    output_columns["output_tensor"] = output_tensor
                    if self.model_type and is_tensorflow(self.model_type[model_index]):
                        output_columns["activation"] = get_activation(output_tensor)

                M.append(output_columns)

                # After a classification is received, call sml_reset model to advance the internal buffer.
                if model_api != "sml_recognition_run":
                    self.reset_model(model_index=model_index)

            if verbose and i in status:
                print("processed {samples} samples".format(samples=i))

        print("finished processing {samples} samples".format(samples=len(data)))

        return DataFrame(M)

    def run_segment(
        self, data_segment: DataFrame, model_index: int = 0, debug_log: bool = False
    ):
        """
        Add a segment of data to the model as input. Then runs the model on the current segment, skipping the data streaming steps.

        (Warning: using this call will flush the models internal ring buffer. Alternating between this and run_and_score model may produce incorrect results or crashes)

        Args:
            data(array):  Array of timseries data
            model_index(int): Index of the model to run
            debug_log(bool): Prints the debug output from the Knowledge Pack

        Returns:
            Classification results will be 0 if unknown through the classification numbers
            you have. This function returns -1 when a segment hasn't yet been identified.

        """
        if not self._initialized():
            return

        if self._run_with("run_segment") is False:
            return

        if isinstance(data_segment, (DataFrame, Series)):
            data = data_segment.values
        elif isinstance(data_segment, ndarray):
            data = data_segment
        else:
            print("Input data Must be either dataframe or array.")
            return

        data_array = (ctypes.c_int16 * (data.shape[0] * data.shape[1]))()
        length_ctype = ctypes.c_int(data.shape[0])
        nbuffs = ctypes.c_int(data.shape[1])
        model_index_ctype = ctypes.c_int(model_index)

        for col in range(data.shape[1]):
            for index, value in enumerate(data[:, col]):
                data_array[col * data.shape[0] + index] = ctypes.c_int16(value)

        self._add_segment(data_array, length_ctype, nbuffs, model_index_ctype)

        model_index_ctype = ctypes.c_int(model_index)

        if USE_PIPES:
            with pipes() as (out, _):
                ret = self._run_segment(model_index_ctype)
        else:
            ret = self._run_segment(model_index_ctype)

        if debug_log and USE_PIPES:
            print(out.read())

        return ret

    def reset_model(self, model_index: int = 0):
        """
        Advances the model state to so that it is ready for a more input data. Use
        only after classification steps.

        (Note: this does not reset the model to a clean state, only init_model does this)

        Args:
           model_index(int): Index of the model to update
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        self._reset_model(model_index_ctype)

    def flush_model(self, model_index: int = 0):
        """
        Deletes all of the patterns in the database.

        Args:
           model_index(int): Index of the model to update
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        self._flush_model(model_index_ctype)

    def get_model(self, model_index: int = 0):
        """
        Prints the model weights and info for current model if supported by this model type.

        Args:
            model index - number of axes in the data stream

        """

        if not self._initialized():
            return

        model_header = self.get_model_header(model_index)
        model = []

        for index in range(model_header.number_patterns):
            tmp_dict = {}
            pattern = self.get_model_pattern(model_index, index)
            tmp_dict["Category"] = pattern.category
            tmp_dict["Vector"] = [
                pattern.vector[i] for i in range(model_header.pattern_length)
            ]
            tmp_dict["AIF"] = pattern.influence
            tmp_dict["Identifier"] = index
            model.append(tmp_dict)

        return DataFrame(model)

    def get_model_score(self, model_index):

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        with pipes() as (out, _):
            self._print_model_score(model_index_ctype)

        model_scores = []
        for line in out.read().split("\n")[:-1]:
            model_scores.append(json.loads(line))

        cats = [str(x) for x in range(1, len(DataFrame(model_scores).columns) - 1)]

        return DataFrame(model_scores)[["ID", "ERR"] + cats].style.apply(_color, axis=1)

    def get_class_map(self, model_index: int = 0) -> Dict:
        """Returns the class map for the model

        Args:
            model_index (int, optional): The model index. Defaults to 0.

        Returns:
            dict: Class map for the specified model_index
        """

        if not self._initialized():
            return None

        return self.class_map[model_index]

    def get_model_header(self, model_index: int = 0):
        """Returns the model header for a PME model

        Args:
            model_index (int, optional): The model index. Defaults to 0.

        Returns:
            struct_pme_model_header: A struct object containing information about the PME model header
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        pme_model_header = struct_pme_model_header()

        with pipes() as (out, _):
            self._get_model_header(model_index_ctype, pme_model_header)

        return pme_model_header

    def get_model_pattern(self, model_index: int = 0, pattern_index: int = 0):
        """Returns the pattern for a PME model at a particular index

        Args:
            model_index (int, optional): The model index. Defaults to 0.
            pattern_index (int, optional): The index of the pattern in the database to retrieve. Defaults to 0.

        Returns:
            struct_pme_pattern: struct_pme_pattern object
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        pattern_index_ctype = ctypes.c_int(pattern_index)
        pme_pattern = struct_pme_pattern()

        with pipes() as (out, _):
            self._get_model_pattern(model_index_ctype, pattern_index_ctype, pme_pattern)

        return pme_pattern

    def get_segment_data(self, model_index: int = 0) -> List:
        """Returns the segment data currently stored in the models internal ring buffer.

        Args:
            model_index (int, optional): The model index. Defaults to 0.

        Returns:
            list: the sensor data stored in the model ring buffer
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        n_sensors = self._kb_get_model_framelen(model_index_ctype)

        number_samples = self._kb_get_model_sg_length(model_index_ctype)
        index = self._kb_get_sg_start_index(model_index_ctype)

        number_samples_ctype = ctypes.c_int(number_samples)
        index_ctype = ctypes.c_int(index)

        buffer_size = number_samples * n_sensors
        sensor_data_buffer = (ctypes.c_short * buffer_size)()

        self._kb_get_segment_data(
            model_index_ctype, number_samples_ctype, index_ctype, sensor_data_buffer
        )

        sensor_data_buffer = list(sensor_data_buffer)
        sensor_data = []
        for i in range(0, buffer_size, number_samples):
            sensor_data.append(sensor_data_buffer[i : i + number_samples])

        return sensor_data

    def get_model_framelen(self, model_index: int = 0) -> int:
        """Returns the number of channels expected by the model

        Args:
            model_index (int): The model index

        Returns:
            int: The number of input channels used by the model
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        framelen = self._kb_get_model_framelen(model_index_ctype)

        return int(framelen)

    def get_feature_vector(
        self, model_index: int = 0, feature_vector_buffer_size: int = 4096
    ) -> List:
        """Get the current feature vector from the model

        Args:
            model_index (int, optional): The model index. Defaults to 0.
            feature_vector_buffer_size (int, optional): The size of the array to use for getting the feature vector (This should be equal to or greater the size of the feature vector used by the model). Defaults to 4096.

        Returns:
            List: The models current feature vector
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        feature_vector_ctype = (ctypes.c_ubyte * feature_vector_buffer_size)()
        feature_length_ctype = ctypes.c_ubyte(0)

        self._get_feature_vector(
            model_index_ctype, feature_vector_ctype, ctypes.byref(feature_length_ctype)
        )

        return list(feature_vector_ctype[: feature_length_ctype.value])

    def get_segment_length(self, model_index: int = 0) -> int:
        """Get the length of the current segment

        Args:
            model_index (int, optional): The model index. Defaults to 0.

        Returns:
            int: The length of the current segment
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        segment_length = ctypes.c_int(0)

        self._sml_get_segment_length(model_index_ctype, ctypes.byref(segment_length))

        return segment_length.value

    def get_feature_bank_number(self, model_index: int = 0) -> int:
        """Get the number of feature banks this model uses

        Args:
            model_index (int, optional): The model index. Defaults to 0.

        Returns:
            int: The number of feature banks
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        feature_bank_number = ctypes.c_int(0)

        self._sml_get_feature_bank_number(
            model_index_ctype, ctypes.byref(feature_bank_number)
        )

        return feature_bank_number.value

    def set_feature_vector(self, model_index: int, feature_vector: list):
        """Sets the models feature vector to the specified values

        Args:
            model_index (int): The model index
            feature_vector (list): the feature vector
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        feature_vector_array = (ctypes.c_ubyte * len(feature_vector))()

        for index, value in enumerate(feature_vector):
            feature_vector_array[index] = ctypes.c_ubyte(value)

        return self._set_feature_vector(model_index_ctype, feature_vector_array)

    def recognize_feature_vector(self, model_index: int = 0) -> int:
        """Classify the feature vector that is currently in the model buffer

        Args:
            model_index (int, optional): The model index. Defaults to 0.

        Returns:
            int: the classification result
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        return self._recognize_feature_vector(model_index_ctype)

    def get_model_result(self, model_index: int = 0, tensor_size: int = 128):
        """
        Get the model tensor output for the most recent classification

        Args:
            model_index (int): the model index to use
            tensor_size (int): the size of the tensor, should be larger than the largest output_tensor size

        Returns:
            model results object for the specified model
        """
        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        model_result = struct_model_result()
        output_tensor = (ctypes.c_int16 * tensor_size)()
        model_result.output_tensor = output_tensor

        self._kb_get_classification_result_info(
            model_index_ctype, ctypes.cast(ctypes.byref(model_result), ctypes.c_void_p)
        )

        outputs = []
        for i in range(model_result.num_outputs):
            outputs.append(model_result.output_tensor[i])

        return outputs

    def knowledgepack(
        self, model_index: int = 0, distance_mode: int = 0, feature_list=None
    ):
        """
        Turn into a Knowledge Pack model object (only works for PME)
        """
        self.init_model()
        neuron_array = self.get_model(model_index).to_dict("records")
        class_map = self.get_class_map(model_index)

        return Model(
            neuron_array,
            class_map,
            {"distance_mode": distance_mode},
            feature_summary=feature_list,
        )

    def run_and_score_model(self, data_sample, category, model_index=0):
        """
        Runs a PME model and updates the internal model score.
        """

        ret = self.run_model(data_sample, model_index)

        if ret is None:
            return

        if ret >= 0:
            print("Class: {}".format(ret))
            if ret > 0:  # only retrain if greater than 0
                score_ret = self.score_model(model_index, category)
                if score_ret == 0:
                    print("Retraining is not supported for this model.")
            self.reset_model(model_index)

            return self.get_model_score(model_index)

        return -1

    def run_and_score_model_on_segment(self, segment, category, model_index=0):
        """
        Run a PME model and updates the internal scores based on a segment of data.

        (Warning: using this call will flush the models internal ring buffer. Alternating between this and run_and_score model may produce incorrect results or crashes)

        """

        ret = self.run_segment(segment, model_index)

        if ret is None:
            return

        if ret >= 0:
            print("Class: {}".format(ret))
            if ret > 0:  # only retrain if greater than 0
                self.score_model(model_index, category)
            self.reset_model(model_index)

        return self.get_model_score(model_index)


def _color(val):
    color = "white"
    if val.ERR > 0:
        color = "#90ee90"
    elif val.ERR < 0:
        color = "red"
    return ["background-color: %s" % color] * len(val)



import json

import sensiml.base.utility as utility
from sensiml.datamanager.capture import Capture, BackgroundCapture


class CaptureExistsError(Exception):
    """Base class for a Capture exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Captures:
    """Base class for a collection of Captures."""

    def __init__(self, connection, project):
        self._connection = connection
        self._project = project
        self._reserved_metadata_names = ["capture_uuid", "segment_uuid"]
        self._capture_list = {}

    def create_capture(self, filename, filepath, asynchronous=True, capture_info=None):
        """Creates a capture object from the given filename and filepath.

        Args:
            filename (str): desired name of the file on the server
            filepath (str): local path to the file to be uploaded
            asynchronous (bool): Whether to process asynchronously

        Returns:
            capture object

        Raises:
            CaptureExistsError, if the Capture already exists on the server
        """
        capture_dict = {
            "connection": self._connection,
            "project": self._project,
            "filename": filename,
        }

        if self.get_capture_by_filename(filename) is not None:
            raise CaptureExistsError("capture {0} already exists.".format(filename))
        else:
            capture = Capture.initialize_from_dict(capture_dict)
            capture.path = filepath
            capture.capture_info = (
                capture_info if isinstance(capture_info, dict) else {}
            )
            capture.insert(asynchronous=asynchronous)

        return capture

    def build_capture_list(self, force: bool = True):
        """Populates the function_list property from the server."""

        if not force and self._capture_list:
            return self._capture_list

        capture_response = self.get_captures()
        for capture in capture_response:
            self._capture_list[capture.filename] = capture

        return self._capture_list

    def get_capture_by_filename(self, filename, force=False):
        """Gets a capture object from the server using its filename property.

        Args:
            filename (str): the capture's name
            force (bool): rebuild the cached list of files

        Returns:
            capture object or None if it does not exist
        """
        capture_list = self.build_capture_list(force=force)

        return capture_list.get(filename)

    def get_capture_by_uuid(self, uuid):
        """Gets a capture object from the server using its filename property.

        Args:
            filename (str): the capture's name

        Returns:
            capture object or None if it does not exist
        """
        """Calls the REST API to update the capture."""
        url = "project/{0}/capture/{1}/".format(self._project.uuid, uuid)

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        return self._new_capture_from_dict(response_data)

    def get_capture_urls_by_uuid(self, capture_uuids, expires_in=100):

        url = f"project/{self._project.uuid}/capture-files/"
        data = {"capture_uuids": capture_uuids, "expires_in": expires_in}

        response = self._connection.request("post", url, json=data)
        response_data, err = utility.check_server_response(response)

        return response_data

    def get_capture_urls_by_name(self, capture_list, expires_in=100):

        capture_uuids = [
            x.uuid
            for key, x in self.build_capture_list().items()
            if key in capture_list
        ]

        return self.get_capture_urls_by_uuid(capture_uuids, expires_in=expires_in)

    def __getitem__(self, key):
        if type(key) == str:
            return self.get_capture_by_filename(key)
        else:
            return self.get_captures()[key]

    def get_statistics(self):
        """Gets all capture statistics for the project.

        Returns:
            DataFrame of capture statistics
        """
        url = "project/{0}/statistics/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            data, error_report = utility.make_statistics_table(response_data)
            return data, error_report
        else:
            return None

    def get_metadata_names_and_values(self):
        """Gets all the metadata names and possible values for a project.

        Returns:
            list(dict) containing metadata names and values
        """
        url = "project/{0}/metadata/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            metadata_values = [
                {
                    "name": item["name"],
                    "values": [i["value"] for i in item["label_values"]],
                }
                for item in response_data
            ]
            return metadata_values

    def get_metadata_names(self):
        """Gets all the metadata names within a project.

        Returns:
            list of metadata names
        """
        url = "project/{0}/metadata/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return [
                item["name"] for item in response_data
            ] + self._reserved_metadata_names

    def get_captures_by_metadata(self, key, value):
        """Gets captures by existing metadata key-values.

        Args:
            key (str): the name of the metadata item
            value (str, int, or float): the value to search for

        Returns:
            list of captures that have the desired metadata key-value pair
        """
        url = "project/{0}/capture/metadata/[{1}]=[{2}]/".format(
            self._project.uuid, key, value
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            captures = []
            for capture_params in response_data:
                captures.append(self._new_capture_from_dict(capture_params))
            return captures

    def get_label_names_and_values(self):
        """Gets all the label names and possible values for a project.

        Returns:
            list(dict) containing metadata names and values
        """
        url = "project/{0}/label/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            metadata_values = [
                {
                    "name": item["name"],
                    "values": [i["value"] for i in item["label_values"]],
                }
                for item in response_data
            ]
            return metadata_values

    def get_label_names(self):
        """Gets all the label names within a project.

        Returns:
            list of label names
        """
        url = "project/{0}/label/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return [item["name"] for item in response_data]

    def _new_capture_from_dict(self, capture_dict):
        """Creates a new capture using the dictionary.

        Args:
            capture_dict (dict): dictionary of capture properties

        Returns:
            capture object
        """
        capture_dict.update({"connection": self._connection, "project": self._project})
        return Capture.initialize_from_dict(capture_dict)

    def get_captures(self):
        """Gets all captures from the server.

        Returns:
            list of captures for the project
        """
        # Query the server and get the json
        url = "project/{0}/capture/".format(self._project.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        # Populate each capture from the server
        captures = []
        if err is False:
            for capture_params in response_data:
                captures.append(self._new_capture_from_dict(capture_params))

        return captures


################################################
class BackgroundCaptures:
    """Base class for a collection of Background Captures."""

    def __init__(self, connection):
        self._connection = connection

    def get_background_captures(self, return_dict=False):
        """Gets all captures from the server.

        Returns:
        list of all available background captures
        """

        # Query the server and get the json
        url = "background-capture/"
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        # Populate each capture from the server
        captures = []
        if err is False:
            for capture_params in response_data:
                capture_params["filename"] = capture_params["name"]
                captures.append(
                    BackgroundCapture(
                        self._connection,
                        **capture_params,
                    )
                )

        return captures

    def get_background_capture_urls_by_uuid(self, capture_uuids, expires_in=100):

        url = f"background-capture-files/"
        data = {"background_capture_uuids": capture_uuids, "expires_in": expires_in}

        response = self._connection.request("post", url, json=data)
        response_data, err = utility.check_server_response(response)

        return response_data

    ################################################



import json
import os


def find_change_index(y, label):
    for index, i in enumerate(y):
        if i != label:
            return index

    return len(y)


def to_dcli(
    df,
    label_column,
    filename,
    session_name=None,
    outdir="dcli_export",
    sensor_columns=None,
    video_name=None,
    metadata=None,
    exclude_labels=None,
    export_dcli=True,
    export_csv=True,
    dcli_name="export.dcli",
    verbose=True,
):
    """
    This is used to convert a CSV file into a dcli file that can be imported into the DCL

    The expected format of a CSV file for this converter is a CSV containing all of the
    sensor data as well as a column that specifies which label each row corresponds to

    Sensor1,Sensor2,...SensorN,Label
    13,134,...,146,Running
    42,124,...,123,Running
    13,342,...,124,Running
    12,134,...,123,Running
    15,121,...,124,Walking
    19,134,...,134,Walking
    .
    .
    .
    19,134,...,134,Walking


    args:
     df (dataframe)
     label_column: the label column to use for adding segments
     filename: the name of the file
     session_name: The session name to use when creating the session for import
     outdir: the directory to write the exported data to
     sensor_columns (list): the column in the DataFrame that are sensor columns, if None use all columns except the label column
     video_name (str): path to the video associated with this file
     metadata (list): metadata already formatted in the dcli format
     exclude_labels (list): labels to exclude when creating this file
     export_dcli (bool): creates the dcli file in the outdir folder
     export_csv (bool): copies the dataframe of sensor data to the the outdir folder
     dcli_name (str): name of the .dcli file to create



    """
    start = 0
    end = 0
    segments = []
    labels = df[label_column].tolist()
    while end < len(labels):
        current_label = labels[end]
        start = labels[end:].index(current_label) + end
        end = find_change_index(labels[start:], current_label) + start

        if start < 0:
            seg_start = 0
        else:
            seg_start = start

        if end > len(labels):
            seg_end = len(labels) - 1
        else:
            seg_end = end - 1

        if exclude_labels and current_label in exclude_labels:
            continue

        segments.append(
            {
                "name": "Label",
                "value": current_label,
                "start": seg_start,
                "end": seg_end,
            }
        )

    if verbose:
        print("Number of segments found", len(segments))
        print("Type of segments found", df[label_column].unique())

    if dcli_name is None:
        dcli_name = "export.dcli"

    if outdir is not None:
        if not os.path.exists(outdir):
            os.mkdir(outdir)
        file_path = os.path.join(outdir, filename)
        dcli_path = os.path.join(outdir, dcli_name)

    else:
        file_path = filename
        dcli_path = dcli_name

    if session_name is None:
        session_name = "Session1"

    dcli_capture = {
        "file_name": filename,
        "sessions": [{"session_name": session_name, "segments": segments}],
    }

    if video_name:
        dcli_capture.update({"videos": [{"video_path": video_name}]})

    if metadata:
        dcli_capture.update(metadata)

    if export_csv:
        if sensor_columns:
            df[sensor_columns].to_csv(file_path, index=None)
        else:
            df[[col for col in df.columns if col != label_column]].to_csv(
                file_path, index=None
            )
        if verbose:
            print("writing csv files to folder '{}'".format(outdir))
    if export_dcli:
        json.dump([dcli_capture], open(dcli_path, "w"))
        if verbose:
            print("writing dcli files to folder '{}'".format(outdir))

    return dcli_capture



import os
import json
from pandas import DataFrame, Series

from sensiml.base import utility


class ArchitectureDescription(object):
    def __init__(self, name=""):
        self._name = name

    @property
    def name(self):
        return self._name


class CompilerDescription(object):
    """Base class for compiler object"""

    def __init__(self):
        self._compiler_version = ""
        self._name = ""
        self._uuid = ""

    @property
    def name(self):
        return self._name

    @property
    def uuid(self):
        return self._uuid

    @property
    def compiler_version(self):
        return self._compiler_version

    def initialize_from_dict(self, init_dict):
        self._uuid = init_dict.get("uuid", "")
        self._compiler_version = init_dict.get("compiler_version", "")
        self._name = init_dict.get("name", "")

    def __dict__(self):
        ret = {
            "Id": self.uuid,
            "Name": self.name,
            "Version": self.compiler_version,
        }
        return ret

    def __call__(self):
        pd_dict = self.__dict__()
        pseries = Series(pd_dict, index=pd_dict.keys())
        df = DataFrame()
        df = df.append(pseries, ignore_index=True)
        return df


class ProcessorDescription(object):
    """Base Class for Processor object"""

    def __init__(self):
        self._uuid = ""
        self._architecture = ArchitectureDescription()
        self._display_name = ""
        self._manufacturer = ""
        self._float_options = dict()
        self._profiling_enabled = False

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True)

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def display_name(self):
        return self._display_name

    @display_name.setter
    def display_name(self, value):
        self._display_name = value

    @property
    def manufacturer(self):
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, value):
        self._manufacturer = value

    @property
    def float_options(self):
        return self._float_options

    @float_options.setter
    def float_options(self, value):
        self._float_options = value

    @property
    def profiling_enabled(self):
        return self._profiling_enabled

    @profiling_enabled.setter
    def profiling_enabled(self, value):
        self._profiling_enabled = value

    def initialize_from_dict(self, init_dict):
        self._uuid = init_dict.get("uuid", "")
        self.float_options = init_dict.get("float_options", list)
        self.manufacturer = init_dict.get("manufacturer", "")
        self.display_name = init_dict.get("display_name", "")
        self.architecture = ArchitectureDescription(
            name=init_dict.get("architecture", "")
        )
        self.profiling_enabled = init_dict.get("profiling_enabled", False)

    def __dict__(self):
        ret = {
            "Id": self.uuid,
            "Name": self.display_name,
            "Manufacturer": self.manufacturer,
            "Float Options": self.float_options,
            "Architecture": self.architecture,
            "Profiling Enabled": self.profiling_enabled,
        }
        return ret

    def __call__(self):
        pd_dict = self.__dict__()
        pseries = Series(pd_dict, index=pd_dict.keys())
        df = DataFrame()
        df = df.append(pseries, ignore_index=True)
        return df


class ClientPlatformDescription(object):
    """Base class for a PlatformDescription object"""

    def __init__(self, connection):
        self._connection = connection
        self._uuid = ""
        self._name = ""
        self._hardware_accelerators = dict()
        self._can_build_binary = False
        self._processors = list()
        self._platform_versions = list()
        self._description = ""
        self._supported_outputs = dict()
        self._supported_compilers = list()
        self._supported_source_drivers = dict()
        self._applications = dict()
        self._default_selections = dict()

    @property
    def name(self):
        return self._name

    @property
    def manufacturer(self):
        return self._manufacturer

    @property
    def uuid(self):
        return self._uuid

    @property
    def id(self):
        return self._uuid

    @property
    def hardware_accelerators(self):
        return self._hardware_accelerators

    @property
    def can_build_binary(self):
        return self._can_build_binary

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def supported_compilers(self):
        return self._supported_compilers

    @property
    def supported_outputs(self):
        return self._supported_outputs

    @property
    def supported_source_drivers(self):
        return self._supported_source_drivers

    @property
    def processors(self):
        return self._processors

    @property
    def default_selections(self):
        return self._default_selections

    @property
    def platform_versions(self):
        return self._platform_versions

    @property
    def applications(self):
        return self._applications

    def get_by_processor_by_uuid(self, uuid):
        for p in self._processors:
            if str(p.uuid) == str(uuid):
                return p
        return None

    def get_config(
        self,
        test_data="",
        debug=False,
        target_processor=None,
        target_compiler=None,
        target_application=None,
        float_options=None,
        platform_version="",
    ):
        """Generates a default configuration dictionary for downloading a knowledgepack using this platform programmatically

        Args:
            test_data (str): name of test data file to load onto knowledge pack,
            debug (bool): Build option for knowledge pack. Debug mode will have extra printouts to help with debugging issues.

        """

        if self.supported_outputs:
            outputs = self.supported_outputs[0]
        else:
            outputs = ["serial"]

        config = {
            "target_platform": self.uuid,
            "test_data": test_data,
            "debug": debug,
            "output_options": outputs,
        }

        if target_application is None:
            config["application"] = list(self.applications.keys())[0]
        else:
            if target_application not in self.applications.keys():
                raise Exception("Invalid application")
            config["application"] = target_application

        if target_processor is None:
            config["target_processor"] = self.processors[0].uuid
        else:
            if sum([target_processor == x.uuid for x in self.processors]) != 1:
                raise Exception("Invalid target process")
            config["target_processor"] = target_processor

        if target_compiler is None:
            config["target_compiler"] = self.supported_compilers[0].uuid
        else:
            if sum([target_compiler == x.uuid for x in self.supported_compilers]) != 1:
                raise Exception("Invalid target compiler")

            config["target_compiler"] = target_compiler

        config["float_options"] = self.get_by_processor_by_uuid(
            config["target_processor"]
        ).float_options.get(self.default_selections.get("float", ""), "")

        if platform_version:
            if platform_version not in self.platform_versions:
                raise Exception("Invalid Platform Version")
        else:
            if self.platform_versions:
                platform_version = self.platform_versions[0]
        config["selected_platform_version"] = platform_version

        return config

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""
        url = "platforms/{0}/".format(self.id)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self._uuid = response_data["uuid"]
            self._name = response_data["name"]
            self._hardware_accelerators = response_data["hardware_accelerators"]
            self._platform_versions = response_data["platform_versions"]
            self._description = response_data["description"]
            self._can_build_binary = response_data["can_build_binary"]
            self._supported_source_drivers = response_data["supported_source_drivers"]
            self._applications = response_data["applications"]
            self._default_selections = response_data.get("default_selections", {})
            self._processors = list()
            for processor in response_data.get("processors", []):
                test = set(self._processors)
                p = ProcessorDescription()
                p.initialize_from_dict(processor)
                if p not in test:
                    self._processors.append(p)

            self._supported_compilers = list()
            for compiler in response_data.get("supported_compilers", []):
                test = set(self._supported_compilers)
                c = CompilerDescription()
                c.initialize_from_dict(compiler)
                if c not in test:
                    self._supported_compilers.append(c)

    def initialize_from_dict(self, input_dictionary):
        """Populates a single transform object from a dictionary of properties from the server."""
        self._uuid = input_dictionary["uuid"]
        self._name = input_dictionary["name"]
        self._hardware_accelerators = input_dictionary["hardware_accelerators"]
        self._platform_versions = input_dictionary["platform_versions"]
        self._description = input_dictionary["description"]
        self._can_build_binary = input_dictionary["can_build_binary"]
        self._applications = input_dictionary["applications"]
        self._default_selections = input_dictionary.get("default_selections", {})
        self._supported_source_drivers = input_dictionary.get(
            "supported_source_drivers",
            {
                "Motion": [200, 100, 50, 25],
                "Audio": [16000],
                "Default": ["Motion", 100],
                "Custom": [],
            },
        )
        for processor in input_dictionary.get("processors", list()):
            proc = ProcessorDescription()
            proc.initialize_from_dict(processor)
            self._processors.append(proc)

        for compiler in input_dictionary.get("supported_compilers", list()):
            comp = CompilerDescription()
            comp.initialize_from_dict(compiler)
            self._supported_compilers.append(comp)

    def __dict__(self):
        def _get_unique_dict_list(in_dict):
            maps = map(dict, set(tuple(sorted(d.__dict__().items())) for d in in_dict))
            inner_ret = list()
            for m in maps:
                inner_ret.append(dict(m))
            return inner_ret

        ret = {
            "Id": self.uuid,
            "Name": self.name,
            "Hardware Accelerated": "Yes"
            if len(self.hardware_accelerators.items()) > 0
            else "No",
            "Processors Supported": self.processors,
            "Compilers Supported": self.supported_compilers,
            "Description": self.description,
        }
        return ret

    def __call__(self):
        pd_dict = self.__dict__()
        pseries = Series(pd_dict, index=pd_dict.keys())
        df = DataFrame()
        df = df.append(pseries, ignore_index=True)
        return df.drop(labels=["Id"], axis=1)



import os
import json
from sensiml.base import utility
from sensiml.datamanager.base import Base, BaseSet


class CustomFunction(Base):
    """Base class for a transform object"""

    _uuid = ""
    _type = None
    _name = None
    _description = ""
    _input_contract = ""
    _output_contract = ""
    _subtype = ""
    _path = ""
    _unit_tests = None
    _logs = ""
    _task_result = ""
    _c_function_name = ""
    _library_pack = ""
    _automl_available = False

    _fields = [
        "uuid",
        "name",
        "transform_type",
        "description",
        "input_contract",
        "output_contract",
        "subtype",
        "unit_tests",
        "logs",
        "task_result",
        "c_function_name",
        "library_pack",
        "task_status",
        "automl_available",
    ]

    _read_only_fields = [
        "uuid",
        "created_at",
        "last_modified",
        "logs",
        "task_result",
        "task_status",
        "automl_available",
    ]

    _field_map = {"transform_type": "type"}

    def __init__(self, connection, uuid=None):
        self._connection = connection
        if uuid:
            self.uuid = uuid
            self.refresh()

    @property
    def base_url(self):
        return "custom-transform/"

    @property
    def detail_url(self):
        return "custom-transform/{uuid}/".format(uuid=self.uuid)

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def subtype(self):
        return self._subtype

    @subtype.setter
    def subtype(self, value):
        self._subtype = value

    @property
    def transform_type(self):
        return "Feature Generator"

    @transform_type.setter
    def transform_type(self, value):
        self._transform_type = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def c_function_name(self):
        return self._c_function_name

    @c_function_name.setter
    def c_function_name(self, value):
        self._c_function_name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def input_contract(self):
        return json.dumps(self._input_contract)

    @input_contract.setter
    def input_contract(self, value):
        self._input_contract = value

    @property
    def output_contract(self):
        return json.dumps(self._output_contract)

    @output_contract.setter
    def output_contract(self, value):
        self._output_contract = value

    @property
    def unit_tests(self):
        return json.dumps(self._unit_tests)

    @unit_tests.setter
    def unit_tests(self, value):
        self._unit_tests = value

    @property
    def logs(self):
        print(self._logs)

    @logs.setter
    def logs(self, value):
        self._logs = value

    @property
    def task_result(self):
        return self._task_result

    @task_result.setter
    def task_result(self, value):
        self._task_result = value

    @property
    def library_pack(self):
        return self._library_pack

    @library_pack.setter
    def library_pack(self, value):
        self._library_pack = value

    @property
    def automl_available(self):
        return self._automl_available

    @automl_available.setter
    def automl_available(self, value):
        self._automl_available = value

    def refresh(self):
        """Calls the REST API and populates the local object properties from the server."""

        response = self._connection.request("get", self.detail_url)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def delete(self):
        """Calls the REST API and populates the local object properties from the server."""

        response = self._connection.request("delete", self.detail_url)

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def insert(self, path):
        """Calls the REST API to insert a new object."""

        data = self._to_representation()

        response = self._connection.file_request(self.base_url, path, data, "rb")

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response

    def update(self, path):
        """Calls the REST API to insert a new object."""

        data = self._to_representation()

        data.pop("library_pack")

        response = self._connection.file_request(
            self.detail_url, path, data, "rb", "put"
        )

        response_data, err = utility.check_server_response(response)

        if err is False:
            self.initialize_from_dict(response_data)

        return response


class CustomFunctionSet(BaseSet):
    def __init__(self, connection, initialize_set=True):
        """Initialize a custom transform set object.

        Args:
            connection
        """
        self._connection = connection
        self._set = None
        self._objclass = CustomFunction
        self._attr_key = "uuid"

        if initialize_set:
            self.refresh()

    @property
    def library_packs(self):
        return self.objs

    @property
    def get_set_url(self):
        return "custom-transform/"

    def _new_obj_from_dict(self, data):
        """Creates a new object from the response data from the server.

        Args:
            data (dict): contains properties of the object

        Returns:
            obj of type _objclass

        """
        obj = self._objclass(self._connection)
        obj.initialize_from_dict(data)
        return obj



from sensiml.dclproj.csv_to_dcli import to_dcli
from sensiml.dclproj.datasegments import (
    audacity_to_datasegments,
    segment_list_to_datasegments,
    DataSegment,
    DataSegments,
)
from sensiml.dclproj.confusion_matrix import ConfusionMatrix
from sensiml.dclproj.dclproj import DCLProject
from sensiml.dclproj.vizualization import plot_threshold_space


__all__ = [
    "DCLProject",
    "to_dcli",
    "segment_list_to_datasegments",
    "audacity_to_datasegments",
    "DataSegment",
    "DataSegments",
    "ConfusionMatrix",
    "plot_threshold_space",
]



import json
import logging
import os
import re
import sys
import urllib
import warnings
from time import sleep
from uuid import UUID

import numpy as np
import pandas as pd
from numpy import int64
from pandas import DataFrame

import sensiml.base.utility as utility
from sensiml.datamanager.clientplatformdescription import ClientPlatformDescription
from sensiml.datamanager.confusion_matrix import (
    ConfusionMatrix,
    ConfusionMatrixException,
)
from sensiml.datamanager.deviceconfig import DeviceConfig
from sensiml.datamanager.featurefile import FeatureFile
from sensiml.method_calls.functioncall import FunctionCall
from sensiml.visualize.knowledgepack_visualize_mixin import VisualizeMixin

logger = logging.getLogger(__name__)


def render_confusion_matrix(cm):

    GT_txt = "GroundTruth_Total"
    keys = [k for k, v in cm.items()]
    gt = [cm[k][GT_txt] for k in keys]
    df = pd.DataFrame.from_dict(
        {k: [cm[k][_] for _ in keys] for k in keys}, orient="index"
    )
    df.columns = keys
    df[GT_txt] = gt

    return df[keys + [GT_txt]]


class KnowledgePackException(Exception):
    """Base exception for KnowledgePacks"""


class KnowledgePack(VisualizeMixin):
    """Base class for a KnowledgePack"""

    def __init__(self, connection, project_uuid, sandbox_uuid=""):
        """Initializes a KnowledgePack instance

        Args:
            project (Project): the parent project of the KnowledgePack
            sandbox (Sandbox): the parent sandbox of the KnowledgePack
        """
        self._uuid = ""
        self._connection = connection
        self._project_uuid = project_uuid
        self._sandbox_uuid = sandbox_uuid
        self._feature_file_uuid = None
        self._project_name = ""
        self._sandbox_name = ""
        self._configuration_index = ""
        self._name = ""
        self._model_index = ""
        self._execution_time = ""
        self._neuron_array = ""
        self._model_results = ""
        self._pipeline_summary = ""
        self._query_summary = ""
        self._feature_summary = ""
        self._device_configuration = ""
        self._transform_summary = ""
        self._knowledgepack_summary = ""
        self._knowledgepack_description = ""
        self._sensor_summary = ""
        self._cost_summary = ""
        self._class_map = ""
        self._device_config = DeviceConfig()
        self._feature_file_cache = None
        self._training_metrics = None

    @property
    def uuid(self):
        return self._uuid

    @property
    def sandbox_uuid(self):
        return self._sandbox_uuid

    @property
    def project_uuid(self):
        return self._project_uuid

    @property
    def sandbox_name(self):
        return self._sandbox_name

    @property
    def project_name(self):
        return self._project_name

    @property
    def name(self):
        return self._name

    @property
    def execution_time(self):
        return self._execution_time

    @property
    def neuron_array(self):
        """The model's neuron array"""
        return self._neuron_array

    @property
    def model_parameters(self):
        """The model's parameters"""
        return self._neuron_array

    @property
    def model_configuration(self):
        """Model Configuration"""
        return self.device_configuration

    @property
    def model_results(self):
        """The model results associated with the KnowledgePack (in JSON form)"""
        return self._model_results

    @property
    def training_metrics(self):
        """The training metrics associated with the KnowledgePack"""
        return self.model_parameters.get("training_metrics", None)

    @property
    def pipeline_summary(self):
        """A summary specification of the pipeline which created the KnowledgePack"""
        return self._pipeline_summary

    @property
    def query_summary(self):
        """A summary specification of the query used by the pipeline which created the KnowledgePack"""
        return self._query_summary

    @property
    def feature_summary(self):
        """A summary of the features generated by the KnowledgePack"""
        return self._feature_summary

    @property
    def feature_statistics(self):
        """A summary of the features generated by the KnowledgePack"""
        return self._feature_summary

    @property
    def configuration_index(self):
        return self._configuration_index

    @property
    def model_index(self):
        return self._model_index

    @property
    def device_configuration(self):
        return self._device_configuration

    @property
    def transform_summary(self):
        """A summary of transform parameters used by the KnowledgePack"""
        return self._transform_summary

    @property
    def sensor_summary(self):
        """A summary of sensor streams used by the KnowledgePack"""
        return self._sensor_summary

    @property
    def class_map(self):
        """A summary of the integer classes/categories used by the KnowledgePack and the corresponding application
        categories"""
        return self._class_map

    @property
    def reverse_class_map(self):
        """A summary of the category/integer class used by the KnowledgePack and the corresponding application
        categories"""
        if self._class_map:
            return {v: k for k, v in self._class_map.items()}
        return {}

    @property
    def feature_statistics(self):
        if self._model_results.get("metrics", None) is None:
            return None

        return DataFrame(self._model_results["feature_statistics"]["validation"])

    def feature_statistics_by_metric(self, metric="validation"):
        if self._model_results.get("feature_statistics", None) is None:
            return None
        return DataFrame(self._model_results["feature_statistics"][metric])

    @property
    def confusion_matrix(self):
        if self._model_results.get("metrics", None) is None:
            return ConfusionMatrix(self._model_results)

        return ConfusionMatrix(self._model_results["metrics"]["validation"])

    def confusion_matrix_by_metric(self, metric="validation"):
        if self._model_results.get("metrics", None) is None:
            return ConfusionMatrix(self._model_results)

        return ConfusionMatrix(self._model_results["metrics"][metric])

    @property
    def cost_dict(self):
        """A summary of device costs incurred by the KnowledgePack"""
        return self._cost_summary

    @property
    def knowledgepack_description(self):
        """Description of knowledgepack. It is for the hierarchical model which created by Autosense"""
        return self._knowledgepack_description

    @property
    def knowledgepack_summary(self):
        """A summary of device costs incurred by the KnowledgePack"""
        return self._knowledgepack_summary

    @property
    def cost_report(self):
        """A printed tabular report of the device cost incurred by the KnowledgePack"""
        return self.get_report("cost")

    @property
    def cost_report_json(self):
        """A JSON report of the Knowledge Pack cost summary"""
        return self.get_report("json")

    def cost_resource_summary(self, processor_uuid=None, hardware_accelerator=None):
        """A summary of resources and time needed in a classification from a Knowledge Pack"""
        return self.get_report(
            "resource_summary",
            processor_uuid=processor_uuid,
            hardware_accelerator=hardware_accelerator,
        )

    def get_report(self, report_type, processor_uuid=None, hardware_accelerator=None):
        """Sends a request for a report to the server and returns the result.

        Args:
            report_type (string): string name of report, ex: 'cost'

        Returns:
            (string): string representation of desired report

        """
        url = "project/{0}/knowledgepack/{1}/report/{2}/".format(
            self._project_uuid, self.uuid, report_type
        )

        if report_type == "cost":
            response = self._connection.request(
                "get", url, headers={"Accept": "text/plain"}
            )
            return response.text
        elif report_type == "json":
            response = self._connection.request(
                "get", url, headers={"Accept": "application/json"}
            )
            return response.json()
        elif report_type == "resource_summary":
            params = {}
            if processor_uuid:
                params["processor_uuid"] = processor_uuid
            if hardware_accelerator:
                params["accelerator"] = hardware_accelerator

            url = "project/{0}/knowledgepack/{1}/report/{2}/".format(
                self._project_uuid, self.uuid, report_type
            )

            response = self._connection.request(
                "get", url, params=params, headers={"Accept": "application/json"}
            )
            return response.json()

    def recognize_features(self, data):
        """Sends a single vector of features to the KnowledgePack for recognition.

        Args:
            data (dict): dictionary containing

              - a feature vector in the format 'Vector': [126, 32, 0, ...]
              - 'DesiredResponses' indicating the number of neuron responses to return

        Returns:
            (dict): dictionary containing

            - CategoryVector (list): numerical categories of the neurons that fired
            - MappedCategoryVector (list): original class categories of the neurons that fired
            - NIDVector (list): ID numbers of the neurons that fired
            - DistanceVector (list): distances of the feature vector to each neuron that fired

        """
        url = "project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_features/".format(
            self._project_uuid, self._sandbox_uuid, self.uuid
        )
        response = self._connection.request("post", url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return response_data

    def recognize_signal(
        self,
        capture=None,
        datafile=None,
        test_plan=None,
        stop_step=False,
        segmenter=True,
        platform="emulator",
        get_result=True,
        kb_description=None,
        compare_labels=False,
        renderer=None,
    ):
        """Sends a DataFrame of raw signals to be run through the feature generation pipeline and recognized.

        Args:
            capturefile (str): The name of a file uploaded through the data capture lab
            datafile (str): The name of an uploading datafile
            platform (str): "emulator" or "cloud". The "emulator" will run compiled c code giving device exact results,
                 the cloud runs similarly to training providing more flexibility in returning early results by setting the
                 stop step.
            stop_step (int): for debugging, if you want to stop the pipeline at a particular step, set stop_step
              to its index
            compare_labels (bool): If there are labels for the input dataframe, use them to create a confusion matrix
            segmenter (bool or FunctionCall): to suppress or override the segmentation algorithm in the original
              pipeline, set this to False or a function call of type 'segmenter' (defaults to True)
            lock (bool, True): If True, waits for the result to return before releasing the ipython cell.

        Returns:
            (dict): dictionary of results and summary statistics from the executed pipeline and recognition
        """
        url = "project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_signal/".format(
            self._project_uuid, self._sandbox_uuid, self.uuid
        )

        if capture:
            data_for_recognition = {"capture": capture}
        elif datafile:
            if datafile[-4:] != ".csv":
                datafile = "{}.csv".format(datafile)
            data_for_recognition = {"datafile": datafile}
        elif platform == "emulator" and not test_plan is None:
            data_for_recognition = {"test_plan": test_plan}
        else:
            print(
                "You must specify a capture, featurefile, datafile or test_plan to execute the recognition process."
            )
            return None

        data_for_recognition["test_plan"] = test_plan

        if platform not in ["cloud", "emulator"]:
            print("Platform must either be cloud or emulator.")

        data_for_recognition["stop_step"] = stop_step
        data_for_recognition["platform"] = platform
        data_for_recognition["kb_description"] = json.dumps(kb_description)
        data_for_recognition["compare_labels"] = compare_labels
        if isinstance(segmenter, bool):
            data_for_recognition["segmenter"] = segmenter
        elif isinstance(segmenter, FunctionCall):
            data_for_recognition["segmenter"] = segmenter._to_dict()
        else:
            raise KnowledgePackException(
                "Segmenter input requires a valid segmentation call object or None."
            )

        if renderer:
            renderer.render("Executing signal classification.")

        try:
            response = self._connection.request("post", url, data_for_recognition)
            response_data, err = utility.check_server_response(response)
            print(response_data)
            if err is False:
                if get_result:
                    results = utility.wait_for_pipeline_result(
                        self, lock=True, silent=True, wait_time=5, renderer=renderer
                    )
                    print("Results retrieved.")

                    if (
                        results[0] is not None
                        and type(results[0]) == DataFrame
                        and "ClassificationName" in results[0].columns.values
                    ):
                        class_count = {}
                        for classification in results[0]["ClassificationName"].values:
                            class_count[classification] = list(
                                results[0]["ClassificationName"].values
                            ).count(classification)
                        if results[1] is None:
                            results[1] = {}
                        results[1]["summary"] = {"class_count": class_count}

                        if "confusion_matrix" in results[1]:
                            results[1]["summary"]["confusion_matrix"] = {}
                            results[1]["summary"]["accuracy"] = {}
                            for session in results[1]["confusion_matrix"].columns:
                                cm = results[1]["confusion_matrix"][session]

                                df = sum(
                                    [
                                        render_confusion_matrix(cm.iloc[i])
                                        for i in range(cm.shape[0])
                                    ]
                                )

                                classes = [
                                    col
                                    for col in df.columns.values
                                    if col != "GroundTruth_Total"
                                ]

                                df["Support"] = df.apply(
                                    lambda row: np.nansum(
                                        [row[col] for col in classes]
                                    ),
                                    axis=1,
                                )
                                df.loc["Predicted"] = df.apply(
                                    lambda row: np.nansum(row)
                                ).values
                                diagonals = []
                                for x in df.index:
                                    if not x in ["Predicted", "Pos_Predic(%)"]:
                                        df.at[x, "Sensitivty(%) "] = (
                                            100 * df.at[x, x] / df.at[x, "Support"]
                                        )
                                        df.at["Pos_Predic(%)", x] = (
                                            100 * df.at[x, x] / df.at["Predicted", x]
                                        )
                                        diagonals.append(df.at[x, x])
                                df.at["Pos_Predic(%)", "Sensitivty(%)"] = (
                                    100.0
                                    * np.nansum(diagonals)
                                    / df.at["Predicted", "Support"]
                                )
                                results[1]["summary"]["confusion_matrix"][session] = df[
                                    classes
                                    + ["GroundTruth_Total", "Support", "Sensitivty(%)"]
                                ]
                                results[1]["summary"]["accuracy"][session] = df.at[
                                    "Pos_Predic(%)", "Sensitivty(%)"
                                ]

                    return results[0], results[1]
                else:
                    print(response_data.get("message", ""))
                    return None, None
        except MemoryError:
            warnings.warn(
                "ERROR: Your system has run out of RAM while performing this operation. Try freeing up some RAM or using less data before attempting this again."
            )
            return None

    def stop_recognize_signal(self):
        """Sends a kill signal to a pipeline"""

        url = "project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_signal/".format(
            self._project_uuid, self._sandbox_uuid, self.uuid
        )
        response = self._connection.request("DELETE", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return None

    def _handle_result(self, response_data):
        def fix_index(df):
            """Converts a string index to an int for a DataFrame"""
            df.index = df.index.astype(int64)
            return df.sort_index()

        def parse_classification_results(df_result):
            """Parses the results returned by knowledgepack recognize_signal into a DataFrame when a classification has
            occurred."""

            df_result.index.name = None
            ordered_columns = [
                "DistanceVector",
                "NIDVector",
                "CategoryVector",
                "MappedCategoryVector",
            ] + [
                i
                for i in df_result.columns
                if i
                not in [
                    "DistanceVector",
                    "NIDVector",
                    "CategoryVector",
                    "MappedCategoryVector",
                ]
            ]
            df_result = df_result[
                [col for col in ordered_columns if col in df_result.columns]
            ]

            return df_result

        summary = {}
        for key in response_data:
            if key not in ["results"]:
                summary[key] = DataFrame(response_data[key])

        # Check for a unicode string which can be returned sometimes
        if isinstance(response_data["results"], str):
            return fix_index(DataFrame(json.loads(response_data["results"]))), summary

        # Check for the results vector and parse appropriately
        if isinstance(response_data["results"], dict) and response_data["results"].get(
            "vectors", None
        ):
            df_result = parse_classification_results(
                DataFrame(response_data["results"]["vectors"])
            )
        else:
            df_result = fix_index(DataFrame(response_data["results"]))

        # If there was labeled data we will have a confusion matrix as well
        if "metrics" in response_data["results"]:
            summary["confusion_matrix"] = ConfusionMatrix(
                response_data["results"]["metrics"]
            )

        return df_result, summary

    def retrieve(self, silent=False):
        """Gets the result of a prior asynchronous execution of the sandbox.

        Returns:
            (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox
            (dict): execution summary including execution time and whether cache was used for each
            step; also contains a feature cost table if applicable
        """

        url = "project/{0}/sandbox/{1}/knowledgepack/{2}/recognize_signal/".format(
            self._project_uuid, self._sandbox_uuid, self.uuid
        )
        response = self._connection.request("get", url)
        data, err = utility.check_server_response(response)
        if err is False:
            response_data = data or {}
            return utility.check_pipeline_status(
                response_data, self._handle_result, silent=silent
            )

    def delete(self):
        """Deletes the knowledgepack

        Returns:
            (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox
            (dict): execution summary including execution time and whether cache was used for each
            step; also contains a feature cost table if applicable
        """

        uuids = {self.uuid}

        if self.knowledgepack_description:
            for _, value in self.knowledgepack_description.items():
                uuids.add(value["uuid"])

        for uuid in uuids:
            url = "project/{0}/sandbox/{1}/knowledgepack/{2}/".format(
                self._project_uuid, self._sandbox_uuid, uuid
            )
            response = self._connection.request("delete", url)
            response, err = utility.check_server_response(response)
            if err is False:
                self._name = name
                print("Knowledgepack " + uuid + " deleted.")

                return response

    def rename(self, name):
        return self.save(name)

    def save(self, name):
        """Rename a knowledge pack

        Returns:
            (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox
            (dict): execution summary including execution time and whether cache was used for each
            step; also contains a feature cost table if applicable
        """

        uuid_list = []
        name_list = []
        uuid_list.append(self.uuid)
        name_list.append(name)

        if self.knowledgepack_description:
            for model_name in self.knowledgepack_description.keys():
                uuid_temp = self.knowledgepack_description[model_name]["uuid"]
                if uuid_temp not in uuid_list:
                    uuid_list.append(uuid_temp)
                    name_list.append(name + "_" + model_name)

        for list_index in range(len(uuid_list)):
            url = "project/{0}/sandbox/{1}/knowledgepack/{2}/".format(
                self._project_uuid, self._sandbox_uuid, uuid_list[list_index]
            )

            data = {"name": name_list[list_index]}
            response = self._connection.request("put", url, data)
            response, err = utility.check_server_response(response)
            if err is False:
                self._name = name
                print("Knowledgepack '" + name_list[list_index] + "' updated.")

                return response

    def terminate_download(self):
        url = "project/{0}/sandbox/{1}/knowledgepack/{2}/generate_lib/".format(
            self._project_uuid, self._sandbox_uuid, self.uuid
        )

        response = self._connection.request("delete", url)
        response, err = utility.check_server_response(response)
        if err is False:
            print(response)

            return response

    def download_library(
        self, folder="", run_async=True, platform=None, renderer=None, *args, **kwargs
    ):
        self.download_library_v2(
            folder=folder,
            run_async=run_async,
            platform=platform,
            renderer=renderer,
            *args,
            **kwargs,
        )

    def download_source(
        self, folder="", run_async=True, platform=None, renderer=None, *args, **kwargs
    ):
        self.download_source_v2(
            folder=folder,
            run_async=run_async,
            platform=platform,
            renderer=renderer,
            *args,
            **kwargs,
        )

    def download_binary(
        self, folder="", run_async=True, platform=None, renderer=None, *args, **kwargs
    ):
        self.download_binary_v2(
            folder=folder,
            run_async=run_async,
            platform=platform,
            renderer=renderer,
            *args,
            **kwargs,
        )

    def download_library_v2(
        self, folder="", run_async=True, platform=None, renderer=None, *args, **kwargs
    ):
        """Calls the server to generate static library image based on device config.

        Args:
            folder (str): Folder to save to if not generating a link

        Returns:
            str: Denoting success, or link to file download
        """
        return self._download_v2(
            folder=folder,
            bin_or_lib="lib",
            run_async=run_async,
            platform=platform,
            renderer=renderer,
            *args,
            **kwargs,
        )

    def download_source_v2(
        self, folder="", run_async=True, platform=None, renderer=None, *args, **kwargs
    ):
        """Calls the server to generate static library image based on device config.

        Args:
            folder (str): Folder to save to if not generating a link

        Returns:
            str: Denoting success, or link to file download
        """
        return self._download_v2(
            folder=folder,
            bin_or_lib="source",
            run_async=run_async,
            platform=platform,
            renderer=renderer,
            *args,
            **kwargs,
        )

    def download_binary_v2(
        self, folder="", run_async=True, platform=None, renderer=None, *args, **kwargs
    ):
        """Calls the server to generate full binary image based on device config.

        Args:
            folder (str): Folder to save to if not generating a link

        Returns:
            str: Denoting success, or link to file download
        """
        return self._download_v2(
            folder=folder,
            bin_or_lib="bin",
            run_async=run_async,
            platform=platform,
            renderer=renderer,
            *args,
            **kwargs,
        )

    def get_sandbox_device_config(self):
        url = "project/{0}/sandbox/{1}/device_config/".format(
            self._project_uuid, self._sandbox_uuid
        )
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response, is_octet=True)
        if err is False:
            return response_data.json().get("device_config")

    def _download_v2(
        self,
        folder,
        bin_or_lib,
        run_async=False,
        platform=None,
        renderer=None,
        *args,
        **kwargs,
    ):
        _path = r"{0}".format(folder)
        url = f"project/{self._project_uuid,}/knowledgepack/{ self.uuid}/generate_{bin_or_lib}/v2"
        # config = self._sandbox.device_config.copy() # get a regular dictionary

        if platform is not None and type(platform) is ClientPlatformDescription:
            kwargs["config"] = {"target_platform": platform.id}
        elif platform is not None and type(platform) is not ClientPlatformDescription:
            err_msg = "Platform not properly specified. Specify via client.platforms()"
            if renderer:
                renderer.render(err_msg)
            return err_msg, False

        config = kwargs.pop("config", None)
        if config:
            if "kb_description" in config and isinstance(
                config["kb_description"], dict
            ):
                config["kb_description"] = json.dumps(config["kb_description"])
            for key in kwargs:
                if key in [
                    "debug",
                    "test_data",
                    "sample_rate",
                    "output_options",
                    "debug_level",
                    "profile",
                    "profile_iterations",
                    "target_processor",
                    "target_compiler",
                    "float_options",
                    "extra_build_flags",
                ]:
                    config.update({key: kwargs[key]})
                if key == "kb_description":
                    config.update({key: json.dumps(kwargs[key])})
        else:
            config = self.get_sandbox_device_config()

        print("Generating {} with configuration".format(bin_or_lib))
        for key in config:
            print("{} : {}".format(key, config[key]))

        if config.get("target_platform", None) is None:
            raise KnowledgePackException("Target platform must be specified.")
        config["asynchronous"] = run_async
        response = self._connection.request("post", url, data=config)
        response_data, err = utility.check_server_response(
            response, is_octet=True, renderer=renderer
        )

        msg = "Generating Knowledge Pack"
        count = 0
        while True:
            response = self._connection.request("get", url)
            response_data, err = utility.check_server_response(
                response, is_octet=True, renderer=renderer
            )
            if (
                err
                or "content-disposition" in response_data.headers
                or response_data.json().get("task_state") in ("FAILURE",)
            ):
                break
            sys.stdout.write(".")
            if renderer:
                msg += " ."
                renderer.render(msg)
            sleep(5)
        if "content-disposition" in response_data.headers:
            if kwargs.get("save_as"):
                filename = kwargs.get("save_as", None)
            else:
                filename = re.findall(
                    'filename="([^"]+)"', response_data.headers["content-disposition"]
                )[0]
            if filename is None:
                filename = re.findall(
                    'filename="([^"]+)"', response_data.headers["content-disposition"]
                )[0]
            file_path = os.path.join(_path, filename)
            with open(file_path, "wb") as f:
                for d in response_data.iter_content(2048):
                    f.write(d)

            output_path = os.path.join(os.path.abspath("."), file_path)
            if renderer:
                renderer.render("KnowledgePack saved to\n{0}".format(output_path))

            return output_path, True
        else:
            err_msg = response_data.json().get("task_result")
            print(err_msg)
            if renderer:
                renderer.render(err_msg)
            return err_msg, False

    def get_build_log(self, renderer=None, save_as=None):
        url = "knowledgepack/{0}/build-logs/".format(self.uuid)

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(
            response, is_octet=True, renderer=renderer
        )

        if "content-disposition" in response_data.headers:
            if save_as:
                filename = save_as
            else:
                filename = re.findall(
                    'filename="([^"]+)"', response_data.headers["content-disposition"]
                )[0]

            with open(filename, "wb") as f:
                for d in response_data.iter_content(2048):
                    f.write(d)

            output_path = os.path.join(os.path.abspath("."), filename)
            if renderer:
                renderer.render("Build Log saved to\n{0}".format(output_path))
            else:
                print("Build Log saved to: {0}".format(output_path))

            return output_path, True
        else:
            err_msg = response_data.body
            print(err_msg)
            if renderer:
                renderer.render(err_msg)
            return err_msg, False

    def get_featurefile(self):

        if self._feature_file_cache is not None:
            return self._feature_file_cache

        class DummyProject:
            def __init__(self, uuid):
                self.uuid = uuid

        ff = FeatureFile(
            self._connection,
            project=DummyProject(self.project_uuid),
            uuid=self._feature_file_uuid,
        )
        ff.refresh()
        res = ff.download_json()

        self._feature_file_cache = DataFrame(json.loads(res.json()))

        return self._feature_file_cache

    def create(self):
        """Create a new knowledge pack on the server using the internal data for this model

        Returns:
            KnowledgePack object
        """

        print(
            "Creating Knowledge Packs is currently beta and the API is subject to change."
        )

        data = {
            "project": self.project_uuid,
            "pipeline_summary": self.pipeline_summary,
            "query_summary": self.query_summary,
            "feature_summary": self.feature_summary,
            "class_map": self.class_map,
            "sensor_summary": self.sensor_summary,
            "transform_summary": self.transform_summary,
            "knowledgepack_summary": self.knowledgepack_summary,
            "sandbox": self.sandbox_uuid,
            "name": self.name,
            "model_parameters": self.model_parameters,
            "model_configuration": self.model_configuration,
        }

        url = f"project/{self.project_uuid}/knowledgepack/"
        response = self._connection.request("post", url, json=data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            kp = KnowledgePack(
                self._connection,
                response_data.get("project_uuid"),
                response_data.get("sandbox_uuid"),
            )
            kp.initialize_from_dict(response_data)
            return kp

        return response_data

    def export(self):
        """Export a Knowledge Pack Model

        Returns:
            Knowledge Pack Export Dict
        """

        url = f"knowledgepack/{self.uuid}/export/"
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return response_data

        return response

    def initialize_from_dict(self, init_dict):
        self._init_dict = init_dict
        self._uuid = init_dict["uuid"]
        self._name = init_dict["name"]
        self._configuration_index = init_dict.get("configuration_index", None)
        self._model_index = init_dict.get("model_index", None)
        self._execution_time = init_dict.get("execution_time", None)

        self._class_map = init_dict["class_map"]
        self._pipeline_summary = init_dict["pipeline_summary"]
        self._query_summary = init_dict["query_summary"]
        self._feature_summary = init_dict["feature_summary"]
        self._transform_summary = init_dict["transform_summary"]
        self._sensor_summary = init_dict["sensor_summary"]

        self._neuron_array = (
            init_dict.get("neuron_array")
            if init_dict.get("neuron_array")
            else init_dict.get("model_parameters")
        )
        self._device_configuration = (
            init_dict.get("device_configuration")
            if init_dict.get("device_configuration")
            else init_dict.get("model_configuration")
        )

        self._model_results = init_dict.get("model_results", None)

        self._cost_summary = init_dict.get("cost_summary", None)
        self._project_name = init_dict.get("project_name")
        self._sandbox_name = init_dict.get("sandbox_name")
        self._feature_file_uuid = init_dict.get("feature_file_uuid")
        self._knowledgepack_summary = init_dict.get("knowledgepack_summary", None)
        self._knowledgepack_description = init_dict.get(
            "knowledgepack_description", None
        )
        self._training_metrics = init_dict.get("training_metrics", None)


def get_knowledgepack(kp_uuid, connection, get_flatbuffer=False):
    """Gets a KnowledgePack by uuid.

    Returns:
        a KnowledgePack instance, list of instances, or None
    """
    url = "knowledgepack/{0}/".format(kp_uuid)
    response = connection.request("get", url, params={"tflite": get_flatbuffer})
    response_data, err = utility.check_server_response(response)
    if err is False:
        kp = KnowledgePack(
            connection,
            response_data.get("project_uuid")
            if response_data.get("project_uuid") is not None
            else response_data.get("project"),
            response_data.get("sandbox_uuid")
            if response_data.get("sandbox_uuid") is not None
            else response_data.get("sandbox"),
        )
        kp.initialize_from_dict(response_data)
        return kp


def delete_knowledgepack(kp_uuid, connection):
    """Gets a KnowledgePack by uuid.

    Returns:
        a KnowledgePack instance, list of instances, or None
    """
    url = "knowledgepack/{0}/".format(kp_uuid)
    response = connection.request("delete", url)


def get_knowledgepacks(connection):
    """Gets the KnowledgePack(s) created by the sandbox.

    Returns:
        a KnowledgePack instance, list of instances, or None
    """
    url = "knowledgepack/"
    response = connection.request("get", url)
    response_data, err = utility.check_server_response(response)
    if err is False:
        return DataFrame(response_data)
